"""EEG Data Viewer"""
import sys
from functools import partial
from queue import Queue
from typing import Callable, Dict, List, Optional, Tuple

import matplotlib.ticker as ticker
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from PyQt5.QtCore import Qt, QTimer  # pylint: disable=no-name-in-module
# pylint: disable=no-name-in-module
from PyQt5.QtWidgets import (QApplication, QCheckBox, QComboBox, QHBoxLayout,
                             QLabel, QPushButton, QSpinBox, QVBoxLayout,
                             QWidget)

from bcipy.acquisition.devices import DeviceSpec
from bcipy.acquisition.util import StoppableProcess
from bcipy.gui.main import static_text_control
from bcipy.gui.viewer.data_source.data_source import QueueDataSource
from bcipy.gui.viewer.data_source.file_streamer import FileStreamer
from bcipy.gui.viewer.data_source.lsl_data_source import LslDataSource
from bcipy.gui.viewer.ring_buffer import RingBuffer
from bcipy.helpers.parameters import DEFAULT_PARAMETERS_PATH, Parameters
from bcipy.helpers.raw_data import settings
from bcipy.signal.process.transform import Downsample, get_default_transform


def filters(
        sample_rate_hz: float, parameters: Parameters
) -> Dict[str, Callable[[np.ndarray, Optional[int]], Tuple[np.ndarray, int]]]:
    """Returns a dict of filters that can be used"""
    return {
        'downsample':
        Downsample(parameters['down_sampling_rate']),
        'default_transform':
        get_default_transform(
            sample_rate_hz=sample_rate_hz,
            notch_freq_hz=parameters['notch_filter_frequency'],
            bandpass_low=parameters['filter_low'],
            bandpass_high=parameters['filter_high'],
            bandpass_order=parameters['filter_order'],
            downsample_factor=parameters['down_sampling_rate'])
    }


def active_indices(all_channels, removed_channels) -> List[int]:
    """List of indices of all channels which will be displayed. By default
    filters out TRG channels and any channels marked as
    removed_channels.

    Parameters
    ----------
    - all_channels : list of channel names
    - removed_channels : list of channel names to exclude

    Returns
    -------
    list of indices for channel names in use.
    """

    return [
        i for i in range(len(all_channels))
        if all_channels[i] not in removed_channels and
        'TRG' not in all_channels[i]
    ]


def init_buffer(samples_per_second: int, seconds: int,
                channels: List[str]) -> RingBuffer:
    """Initialize the underlying RingBuffer by pre-allocating empty
    values. Buffer size is determined by the sample frequency and the
    number of seconds to display.

    Parameters
    ----------
    - samples_per_second : sample rate
    - seconds : number of seconds of data to store
    - channels : list of channel names
    """

    buf_size = int(samples_per_second * seconds)
    empty_val = [0.0 for _x in channels]
    return RingBuffer(buf_size, pre_allocated=True, empty_value=empty_val)


class FixedHeightHBox(QWidget):
    """Container for holding controls for the EEG Viewer. Acts like a
    QHBoxLayout with a fixed height."""

    def __init__(self, height: int = 30):
        super().__init__()
        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.setFixedHeight(height)

    def addWidget(self, widget: QWidget):
        """Add the given widget to the layout"""
        self.layout.addWidget(widget)


class ChannelControls(QWidget):
    """Controls for toggling channels"""

    def __init__(self, font_size: int, channels: List[str],
                 active_channel_indices: List[int],
                 toggle_channel_fn: Callable):
        super().__init__()
        control_stylesheet = f"font-size: {font_size}px;"
        channel_box = QHBoxLayout()
        channel_box.setContentsMargins(0, 0, 0, 0)

        for channel_index in active_channel_indices:
            channel_name = channels[channel_index]
            chkbox = QCheckBox(channel_name)
            chkbox.setChecked(True)
            chkbox.setStyleSheet(control_stylesheet)
            chkbox.toggled.connect(partial(toggle_channel_fn, channel_index))
            channel_box.addWidget(chkbox)
        self.setLayout(channel_box)
        self.setFixedHeight(font_size + 4)


class FixedScaleInput(QWidget):
    """Input for adjusting the fixed scale value"""

    def __init__(self,
                 initial_value: int,
                 on_change_fn: Callable,
                 label: str = 'Fixed scale:',
                 max_value: int = 5000,
                 font_size: int = 11):
        super().__init__()
        stylesheet = f"font-size: {font_size}px;"
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        lbl = QLabel(label)
        lbl.setStyleSheet(stylesheet)
        lbl.setAlignment(Qt.AlignRight)
        lbl.setContentsMargins(0, 2, 0, 0)
        layout.addWidget(lbl)

        self.fixed_scale_input = QSpinBox()
        self.fixed_scale_input.setMaximum(max_value)
        self.fixed_scale_input.setValue(initial_value)
        self.fixed_scale_input.setStyleSheet(stylesheet)
        self.fixed_scale_input.setAlignment(Qt.AlignLeft)
        self.fixed_scale_input.valueChanged.connect(on_change_fn)
        self.fixed_scale_input.setFocusPolicy(Qt.ClickFocus)

        layout.addWidget(self.fixed_scale_input)
        self.setLayout(layout)

    def value(self) -> int:
        """Returns the input value"""
        return self.fixed_scale_input.value()


class EEGPanel(QWidget):
    """GUI Frame in which data is plotted. Plots a subplot for every channel.
    Relies on a Timer to retrieve data at a specified interval. Data to be
    displayed is retrieved from a provided DataSource.

    Parameters:
    -----------
    - data_source : object that implements the viewer DataSource interface.
    - device_spec : metadata about the data.
    - parameters : configuration for filters, etc.
    - seconds : how many seconds worth of data to display.
    - refresh : time in milliseconds; how often to refresh the plots
    - y_scale : max y-value to use when using a fixed scale for plots
    (autoscale turned off);
    """
    control_font_size = 11
    # space between axis label and tick labels
    yaxis_label_space = 60
    yaxis_label_fontsize = 12
    # fixed width font so we can adjust spacing predictably
    yaxis_tick_font = 'DejaVu Sans Mono'
    yaxis_tick_fontsize = 10

    def __init__(self,
                 data_source,
                 device_spec: DeviceSpec,
                 parameters: Parameters,
                 seconds: int = 5,
                 refresh: int = 500,
                 y_scale=500):
        super().__init__()

        self.data_source = data_source
        self.parameters = parameters
        self.refresh_rate = refresh
        self.samples_per_second = device_spec.sample_rate
        self.records_per_refresh = int(
            (self.refresh_rate / 1000) * self.samples_per_second)

        self.channels = device_spec.channels
        self.removed_channels = ['TRG', 'timestamp']
        self.active_channel_indices = active_indices(self.channels,
                                                     self.removed_channels)

        self.seconds = seconds
        self.downsample_factor = parameters['down_sampling_rate']

        self.filter_options = filters(self.samples_per_second, parameters)
        self.selected_filter = 'downsample'

        self.autoscale = True
        self.y_scale = y_scale

        self.started = False
        self.initUI()

        self.timer = QTimer()
        self.timer.timeout.connect(self.update_plots)

        # The buffer stores raw, unfiltered data.
        self.buffer = init_buffer(self.samples_per_second, self.seconds,
                                  self.channels)
        self.init_data_plots()
        self.axes_changed = False

    # pylint: disable=attribute-defined-outside-init
    def init_canvas(self):
        """Initialize the Figure for drawing plots"""
        self.y_min = -self.y_scale
        self.y_max = self.y_scale

        # figure size is in inches.
        self.figure = Figure(figsize=(12, 9),
                             dpi=72,
                             tight_layout={'pad': 0.0})

        self.axes = self.init_axes()
        self.axes_bounds = self.init_axes_bounds()
        self.canvas = FigureCanvasQTAgg(self.figure)

    # pylint: disable=invalid-name,attribute-defined-outside-init
    def initUI(self):
        """Initialize the UI"""
        vbox = QVBoxLayout()

        self.init_canvas()
        vbox.addWidget(self.canvas)

        # Toolbar
        self.toolbar = QVBoxLayout()

        controls = FixedHeightHBox()

        control_stylesheet = f"font-size: {self.control_font_size}px;"

        # Start/Pause button
        self.start_stop_btn = QPushButton('Pause' if self.started else 'Start',
                                          self)
        self.start_stop_btn.setFixedWidth(80)
        self.start_stop_btn.setStyleSheet(control_stylesheet)
        self.start_stop_btn.clicked.connect(self.toggle_stream)
        controls.addWidget(self.start_stop_btn)

        # Pulldown list of seconds to display
        self.seconds_choices = [2, 5, 10]
        if self.seconds not in self.seconds_choices:
            self.seconds_choices.append(self.seconds)
            self.seconds_choices.sort()

        self.seconds_input = QComboBox()
        self.seconds_input.addItems(
            [str(x) + " seconds" for x in self.seconds_choices])
        self.seconds_input.setCurrentIndex(
            self.seconds_choices.index(self.seconds))
        self.seconds_input.setStyleSheet(control_stylesheet)
        self.seconds_input.currentIndexChanged.connect(self.seconds_handler)
        controls.addWidget(self.seconds_input)

        # Autoscale checkbox
        self.autoscale_checkbox = QCheckBox('Auto-scale')
        self.autoscale_checkbox.setStyleSheet(control_stylesheet)
        self.autoscale_checkbox.setChecked(self.autoscale)
        self.autoscale_checkbox.toggled.connect(self.toggle_autoscale_handler)
        controls.addWidget(self.autoscale_checkbox)

        # Fixed scale input
        self.fixed_scale_input = FixedScaleInput(
            self.y_scale,
            self.fixed_scale_handler,
            label='Fixed scale:',
            font_size=self.control_font_size)
        controls.addWidget(self.fixed_scale_input)

        # Filter checkbox
        self.sigpro_checkbox = QCheckBox('Filtered')
        self.sigpro_checkbox.setStyleSheet(control_stylesheet)
        self.sigpro_checkbox.setChecked(False)
        self.sigpro_checkbox.toggled.connect(self.toggle_filtering_handler)
        controls.addWidget(self.sigpro_checkbox)

        # Filter settings label
        filter_settings = [
            f"Downsample: {self.downsample_factor}",
            f"Notch Freq: {self.parameters['notch_filter_frequency']}",
            f"Low: {self.parameters['filter_low']}",
            f"High: {self.parameters['filter_high']}",
            f"Order: {self.parameters['filter_order']}"
        ]
        self.filter_settings_text = static_text_control(
            self,
            label=f"[{', '.join(filter_settings)}]",
            size=10,
            color='dimgray')
        self.filter_settings_text.setWordWrap(False)
        controls.addWidget(self.filter_settings_text)

        self.toolbar.addWidget(controls)

        channel_box = ChannelControls(self.control_font_size, self.channels,
                                      self.active_channel_indices,
                                      self.toggle_channel)
        self.toolbar.addWidget(channel_box)

        vbox.addLayout(self.toolbar)

        self.setWindowTitle('EEG Viewer')
        self.setLayout(vbox)
        self.show()

    @property
    def filter(self):
        """Returns the current filter"""
        return self.filter_options[self.selected_filter]

    def init_axes(self):
        """Sets configuration for axes. Creates a subplot for every data
        channel and configures the appropriate labels and tick marks."""

        axes = self.figure.subplots(len(self.active_channel_indices),
                                    1,
                                    sharex=True)
        for i, channel in enumerate(self.active_channel_indices):
            ch_name = self.channels[channel]
            axes[i].set_frame_on(False)
            axes[i].set_ylabel(ch_name,
                               rotation=0,
                               labelpad=self.yaxis_label_space,
                               fontsize=self.yaxis_label_fontsize)
            # x-axis shows seconds in 0.5 sec increments
            tick_names = np.arange(0, self.seconds, 0.5)
            ticks = [(self.samples_per_second * sec) / self.downsample_factor
                     for sec in tick_names]
            axes[i].xaxis.set_major_locator(ticker.FixedLocator(ticks))
            axes[i].xaxis.set_major_formatter(
                ticker.FixedFormatter(tick_names))
            axes[i].tick_params(axis='y',
                                which='major',
                                labelsize=self.yaxis_tick_fontsize)
            for tick in axes[i].get_yticklabels():
                tick.set_fontname(self.yaxis_tick_font)
            axes[i].grid()
        return axes

    def reset_axes(self):
        """Clear the data in the GUI."""
        self.figure.clear()
        self.axes = self.init_axes()
        self.axes_bounds = self.init_axes_bounds()
        self.init_data_plots()

    def init_axes_bounds(self):
        """Initial value for ymin, ymax bounds for every axis."""
        return [None] * len(self.axes)

    def toggle_channel(self, channel_index):
        """Remove the provided channel from the display"""
        channel = self.channels[channel_index]
        if self.started:
            self.stop()

        if channel in self.removed_channels:
            self.removed_channels.remove(channel)
        else:
            self.removed_channels.append(channel)
        self.active_channel_indices = active_indices(self.channels,
                                                     self.removed_channels)
        self.axes_changed = True

    def start(self):
        """Start streaming data in the viewer."""
        # update buffer with latest data on (re)start.
        self.start_stop_btn.setText('Pause')
        stylesheet = f'font-size: {self.control_font_size}px;'
        self.start_stop_btn.setStyleSheet(stylesheet)
        self.start_stop_btn.repaint()
        self.started = True

        if self.axes_changed:
            self.axes_changed = False
            self.reset_axes()

        self.update_buffer(fast_forward=True)
        self.timer.start(self.refresh_rate)

    def stop(self):
        """Stop/Pause the viewer."""
        self.start_stop_btn.setText('Start')
        stylesheet = (f'font-size: {self.control_font_size}px;'
                      'color: white; background-color: green;'
                      'border: 1px solid darkgreen;'
                      'border-radius: 4px; padding: 3px;')
        self.start_stop_btn.setStyleSheet(stylesheet)
        self.start_stop_btn.repaint()
        self.started = False
        self.timer.stop()

    def toggle_stream(self):
        """Toggle data streaming"""
        if self.started:
            self.stop()
        else:
            self.start()

    def toggle_filtering_handler(self):
        """Event handler for toggling data filtering."""
        self.with_refresh(self.toggle_filtering)

    def toggle_filtering(self):
        """Toggles data filtering."""
        self.selected_filter = 'default_transform' if self.sigpro_checkbox.isChecked(
        ) else 'downsample'

    def toggle_autoscale_handler(self):
        """Event handler for toggling autoscale"""
        self.with_refresh(self.toggle_autoscale)

    def toggle_autoscale(self):
        """Sets autoscale to checkbox value"""
        self.autoscale = self.autoscale_checkbox.isChecked()

    def seconds_handler(self):
        """Event handler for changing seconds"""
        self.with_refresh(self.update_seconds)

    def update_seconds(self):
        """Set the number of seconds worth of data to display from the
        pulldown list."""
        self.seconds = self.seconds_choices[self.seconds_input.currentIndex()]
        self.buffer = init_buffer(self.samples_per_second, self.seconds,
                                  self.channels)

    def fixed_scale_handler(self):
        """Event handler for updating the fixed scale"""
        self.with_refresh(self.update_fixed_scale)

    def update_fixed_scale(self):
        """Sets the fixed scale value from the input."""
        self.y_scale = abs(self.fixed_scale_input.value())
        self.y_min = -self.y_scale
        self.y_max = self.y_scale
        self.autoscale = False
        self.autoscale_checkbox.setChecked(self.autoscale)
        self.autoscale_checkbox.repaint()

    def with_refresh(self, fn):
        """Pauses streaming, performs the given action, and sets a flag
        indicating that the display axes should be refreshed on restart."""

        if self.started:
            self.stop()
        fn()
        self.axes_changed = True

    @property
    def current_data(self):
        """Returns the data as an np array with a row of floats for each
        displayed channel."""

        # array of 'object'; TRG data may be strings
        data = np.array(self.buffer.data)

        # select only data columns and convert to float
        return np.array(data[:, self.active_channel_indices],
                        dtype='float64').transpose()

    def filtered_data(self):
        """Channel data after applying the current filter.

        Returns the data as an np array with a row of floats for each
        displayed channel.
        """
        channel_data, _fs = self.filter(self.current_data,
                                        self.samples_per_second)
        return channel_data

    @property
    def cursor_x(self):
        """Current cursor position (x-axis), accounting for downsampling."""
        return self.buffer.cur // self.downsample_factor

    def init_data_plots(self):
        """Initialize the data in the viewer."""
        channel_data = self.filtered_data()

        cursor_x = self.cursor_x
        for i, _channel in enumerate(self.active_channel_indices):
            # Initial data plotted will be a list 0.0; this will be updated
            # with real data every tick using the `update_plots` method.
            data = channel_data[i].tolist()
            self.axes[i].set_ybound(lower=self.y_min, upper=self.y_max)
            self.axes[i].plot(data, linewidth=0.8)
            # plot cursor
            self.axes[i].axvline(cursor_x, color='r')

    def update_buffer(self, fast_forward: bool = False) -> None:
        """Update the buffer with latest data from the datasource.
        If the datasource does not have the requested number of
        samples, viewer streaming is stopped."""
        try:
            records = self.data_source.next_n(self.records_per_refresh,
                                              fast_forward=fast_forward)
            for row in records:
                self.buffer.append(row)
        except StopIteration:
            self.stop()
        except BaseException:
            self.stop()

    def update_axes_bounds(self, index: int, data: List[float]) -> bool:
        """Update the upper and lower bounds of the axis at the given index.

        Parameters
        ----------
        - index : index of the axis to update
        - data : current data to be displayed

        Returns
        -------
        bool indicating whether or not the bounds changed
        """
        assert self.axes_bounds, "Axes bounds must be initialized"
        # Maintains a symmetrical scale. Alternatively we could set:
        #   data_min = round(min(data))
        #   data_max = round(max(data))
        data_max = max(abs(round(min(data))), abs(round(max(data))))
        data_min = -data_max
        if self.axes_bounds[index]:
            current_min, current_max = self.axes_bounds[index]
            self.axes_bounds[index] = (min(current_min, data_min),
                                       max(current_max, data_max))
            return (current_min, current_max) != self.axes_bounds[index]
        self.axes_bounds[index] = (data_min, data_max)
        return True

    def update_plots(self):
        """Called by the timer on refresh. Updates the buffer with the latest
        data and refreshes the plots. This is called on every tick."""
        self.update_buffer()
        channel_data = self.filtered_data()

        cursor_x = self.cursor_x
        # plot each channel
        for i, _channel in enumerate(self.active_channel_indices):
            data = channel_data[i].tolist()
            self.axes[i].lines[0].set_ydata(data)
            # cursor line
            self.axes[i].lines[1].set_xdata(cursor_x)
            if self.autoscale:
                bounds_changed = self.update_axes_bounds(i, data)
                data_min, data_max = self.axes_bounds[i]

                self.axes[i].set_ybound(lower=data_min, upper=data_max)

                if bounds_changed:
                    # For ylabels to be aligned consistently, labelpad is re-calculated
                    ch_name = self.channels[_channel]
                    tick_labels = self.axes[i].get_yticks()
                    # Min tick value does not display so index is 1, not 0.
                    pad = self.adjust_padding(data_min, data_max)
                    self.axes[i].set_ylabel(ch_name,
                                            rotation=0,
                                            labelpad=pad,
                                            fontsize=12)
            else:
                self.axes[i].set_ybound(lower=self.y_min, upper=self.y_max)

        self.canvas.draw()

    def adjust_padding(self, data_min: int, data_max: int) -> int:
        """Attempts to keep the channel labels in the same position by adjusting
        the padding between the yaxis label and the yticks."""
        digits_min = len(str(data_min))
        digits_max = len(str(data_max))
        chars = max(digits_min, digits_max)

        # assume at least 2 digits to start.
        baseline_chars = 2
        # Value determined by trial and error; this may change if the tick font
        # or font size is adjusted.
        ytick_digit_width = 7
        return self.yaxis_label_space - \
            ((chars - baseline_chars) * ytick_digit_width)


def lsl_data() -> Tuple[LslDataSource, DeviceSpec, None]:
    """Constructs an LslDataSource, which provides data written to an LSL EEG
    stream."""
    data_source = LslDataSource(stream_type='EEG')
    return (data_source, data_source.device_spec, None)


def file_data(path: str
              ) -> Tuple[QueueDataSource, DeviceSpec, StoppableProcess]:
    """Constructs a QueueDataSource from the contents of the file at the given
    path. Data is written to the datasource at the rate specified in the
    raw_data.csv metadata, so it acts as a live stream.

    Parameters
    ----------
        path - path to the raw_data.csv file
    Returns
    -------
        (QueueDataSource, DeviceInfo, FileStreamer)
            - QueueDataSource can be provided to the EEGFrame
            - DeviceInfo provides the metadata read from the raw_data.csv
            - data is not written to the QueueDataSource until the
                FileStreamer (StoppableProcess) is started.
    """
    # read metadata
    name, freq, channels = settings(path)
    queue = Queue()
    streamer = FileStreamer(path, queue)
    data_source = QueueDataSource(queue)
    device_spec = DeviceSpec(name=name, channels=channels, sample_rate=freq)
    streamer.start()

    return (data_source, device_spec, streamer)


def main(data_file: str,
         seconds: int,
         refresh: int,
         yscale: int,
         display_screen: int = 1,
         parameters: str = DEFAULT_PARAMETERS_PATH):
    """Run the viewer GUI. If a raw_data.csv data_file is provided, data is
    streamed from that, otherwise it will read from an LSLDataStream by
    default.

    Parameters:
    -----------
        data_file - optional, raw_data.csv file to stream.
        seconds - how many seconds worth of data to display.
        downsample_factor - how much the data is downsampled. A factor of 1
            displays the raw data.
        display_screen - monitor in which to display the viewer
            (0 for primary, 1 for secondary)
        parameters - location of parameters.json file with configuration for filters.
    """
    data_source, device_spec, proc = file_data(
        data_file) if data_file else lsl_data()

    app = QApplication(sys.argv)
    panel = EEGPanel(data_source, device_spec,
                     Parameters(parameters, cast_values=True), seconds,
                     refresh, yscale)

    if display_screen == 1 and len(app.screens()) > 1:
        # place frame in the second monitor if one exists
        non_primary_screens = [
            screen for screen in app.screens()
            if screen != app.primaryScreen()
        ]
        display_monitor = non_primary_screens[0]
        monitor = display_monitor.geometry()
    else:
        monitor = app.primaryScreen().geometry()

    # increase height to 90% of monitor height and preserve aspect ratio.
    new_height = int(monitor.height() * 0.9)
    pct_increase = (new_height - panel.height()) / panel.height()
    new_width = panel.width() + int(panel.width() * pct_increase)

    panel.resize(new_width, new_height)
    panel.move(monitor.left(), monitor.top())

    panel.start()

    app_exit = app.exec_()
    if proc:
        proc.stop()
    sys.exit(app_exit)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-f',
                        '--file',
                        help='path to the data file',
                        default=None)
    parser.add_argument('-p',
                        '--parameters',
                        default=DEFAULT_PARAMETERS_PATH,
                        help='Parameter location. Pass as *.json')
    parser.add_argument('-s',
                        '--seconds',
                        help='seconds to display',
                        default=5,
                        type=int)
    parser.add_argument('-r',
                        '--refresh',
                        help='refresh rate in ms',
                        default=500,
                        type=int)
    parser.add_argument('-y', '--yscale', help='yscale', default=500, type=int)
    parser.add_argument('-m',
                        '--monitor',
                        help='display screen (0: primary, 1: secondary)',
                        default=0,
                        type=int)

    args = parser.parse_args()
    main(data_file=args.file,
         seconds=args.seconds,
         refresh=args.refresh,
         yscale=args.yscale,
         display_screen=args.monitor,
         parameters=args.parameters)



from enum import Enum

from psychopy import core, visual

from bcipy.feedback.feedback import Feedback
from bcipy.helpers.clock import Clock
from bcipy.helpers.stimuli import resize_image


class FeedbackType(Enum):
    TEXT = 'TEXT'
    IMAGE = 'IMAGE'


class VisualFeedback(Feedback):
    """Visual Feedback."""

    def __init__(self, display, parameters: dict, clock: Clock):

        # Register Feedback Type
        self.feedback_type = 'Visual Feedback'

        super(VisualFeedback, self).__init__(self.feedback_type)

        # Display Window
        self.display = display

        # Parameters
        self.parameters = parameters
        self.font_stim = self.parameters['feedback_font']
        self.height_stim = self.parameters['feedback_stim_height']

        pos_x = self.parameters['feedback_pos_x']
        pos_y = self.parameters['feedback_pos_y']

        self.pos_stim = (pos_x, pos_y)

        self.feedback_length = self.parameters['feedback_duration']
        self.color = self.parameters['feedback_color']

        # Clock
        self.clock = clock

        self.feedback_timestamp_label = 'visual_feedback'

    def administer(
            self,
            stimulus,
            stimuli_type=FeedbackType.TEXT):
        """Administer.

        Administer visual feedback. Timing information from parameters,
            current feedback given by stimulus.
        """

        stim = self._construct_stimulus(
            stimulus,
            self.pos_stim,
            self.color,
            stimuli_type)

        time = self._show_stimuli(stim)

        core.wait(self.feedback_length)

        return [time]

    def _show_stimuli(self, stimulus) -> None:
        stimulus.draw()
        time = [self.feedback_timestamp_label, self.clock.getTime()]  # TODO: use callback for better precision
        self.display.flip()
        return time

    def _resize_image(self, stimulus, display_size, stimuli_height):
        return resize_image(
            stimulus, display_size, stimuli_height)

    def _construct_stimulus(self, stimulus, pos, fill_color, stimuli_type):
        if stimuli_type == FeedbackType.IMAGE:
            image_stim = visual.ImageStim(
                win=self.display,
                image=stimulus,
                mask=None,
                pos=pos,
                ori=0.0)
            image_stim.size = self._resize_image(
                stimulus, self.display.size, self.height_stim)
            return image_stim
        if stimuli_type == FeedbackType.TEXT:
            return visual.TextStim(
                win=self.display,
                font=self.font_stim,
                text=stimulus,
                height=self.height_stim,
                pos=pos,
                color=fill_color)



"""Data server that streams EEG data over a LabStreamingLayer StreamOutlet
using pylsl."""
import logging
from queue import Empty, Queue
from typing import Generator
import time
import uuid

from pylsl import StreamInfo, StreamOutlet

from bcipy.config import DEFAULT_ENCODING
from bcipy.acquisition.datastream.generator import random_data_generator
from bcipy.acquisition.datastream.producer import Producer
from bcipy.acquisition.devices import DeviceSpec
from bcipy.acquisition.util import StoppableThread

log = logging.getLogger(__name__)

# pylint: disable=too-many-arguments

MARKER_STREAM_NAME = 'TRG_device_stream'


class LslDataServer(StoppableThread):
    """Data server that streams EEG data over a LabStreamingLayer StreamOutlet
    using pylsl. See https://github.com/sccn/labstreaminglayer/wiki.

    In parameters.json, if the fake_data parameter is set to true and the
    device is set to DSI-24, this server will be used to mock data. Alternatively,
    fake_data can be set to false and this module can be run standalone in its
    own python instance.

    Parameters
    ----------
        device_spec : DeviceSpec
            parameters used to configure the server. Should have at least
            a list of channels and the sample frequency.
        generator : optional Generator (see generator.py for options)
            used to generate the data to be served. Uses random_data_generator
            by default.
        include_meta: bool, optional
            if True, writes metadata to the outlet stream.
        add_markers: bool, optional
            if True, creates a the marker channel and streams data to this
            channel at a fixed frequency.
        marker_stream_name: str, optional
            name of the sample marker stream
    """

    def __init__(self, device_spec: DeviceSpec, generator: Generator = None, include_meta: bool = True,
                 add_markers: bool = False, marker_stream_name: str = MARKER_STREAM_NAME):
        super(LslDataServer, self).__init__()

        self.device_spec = device_spec
        self.generator = generator or random_data_generator(channel_count=device_spec.channel_count)

        log.debug("Starting LSL server for device: %s", device_spec.name)
        print(f"Serving: {device_spec}")
        info = StreamInfo(device_spec.name,
                          device_spec.content_type, device_spec.channel_count,
                          device_spec.sample_rate,
                          device_spec.data_type,
                          f'{device_spec.content_type.lower()}_{uuid.uuid1()}')

        if include_meta:
            meta_channels = info.desc().append_child('channels')
            for channel in device_spec.channel_specs:
                channel_node = meta_channels.append_child('channel')
                channel_node.append_child_value('label', channel.name)
                if channel.units:
                    channel_node.append_child_value('unit', channel.units)
                if channel.type:
                    channel_node.append_child_value('type', channel.type)

        self.outlet = StreamOutlet(info)

        self.add_markers = add_markers
        if add_markers:
            # Marker stream (C++ source below):
            # lsl::stream_info marker_info("gUSBamp-"+deviceNumber+"Markers",
            # "Markers",1,0,lsl::cf_string,
            # "gUSBamp_" + boost::lexical_cast<std::string>(deviceNumber) +
            # "_" + boost::lexical_cast<std::string>(serialNumber) +
            # "_markers");
            log.debug("Creating marker stream")
            print(f"Marker channel name: {marker_stream_name}")
            markers_info = StreamInfo(marker_stream_name,
                                      "Markers", 1, 0, 'string',
                                      "uid12345_markers")
            self.markers_outlet = StreamOutlet(markers_info)
        self.started = False

    def stop(self):
        """Stop the thread and cleanup resources."""

        log.debug("[*] Stopping data server")
        super(LslDataServer, self).stop()

        # Allows pylsl to cleanup; The outlet will no longer be discoverable
        # after destruction and all connected inlets will stop delivering data.
        del self.outlet
        self.outlet = None

        if self.add_markers:
            del self.markers_outlet
            self.markers_outlet = None

    def run(self):
        """Main loop of the thread. Continuously streams data to the stream
        outlet at a rate consistent with the sample frequency. May also
        output markers at a different interval."""

        sample_counter = 0
        self.started = True

        data_queue = Queue()
        with Producer(data_queue, generator=self.generator,
                      freq=1 / self.device_spec.sample_rate):
            while self.running():
                sample_counter += 1
                try:
                    sample = data_queue.get(True, 2)
                    self.outlet.push_sample(sample)
                    if self.add_markers and sample_counter % 1000 == 0:
                        self.markers_outlet.push_sample(["1"])
                except (Empty, AttributeError):
                    # outlet.push_sample(sample) may cause an error after
                    # the server has been stopped since the attribute is
                    # deleted in another thread.
                    break

        log.debug("[*] No longer pushing data")


def _settings(filename):
    """Read the daq settings from the given data file"""

    with open(filename, 'r', encoding=DEFAULT_ENCODING) as datafile:
        daq_type = datafile.readline().strip().split(',')[1]
        sample_hz = int(datafile.readline().strip().split(',')[1])
        channels = datafile.readline().strip().split(',')
        return (daq_type, sample_hz, channels)


def await_start(dataserver: LslDataServer, max_wait: float = 2):
    """Blocks until server is started. Raises if max_wait is exceeded before
    server is started.

    Parameters
    ----------
        dataserver - instantiated (unstarted) server on which to wait.
        max_wait - the max number of seconds to wait. After this period if the
            server has not succesfully started an exception is thrown.
    """

    dataserver.start()
    wait = 0
    wait_interval = 0.01
    while not dataserver.started:
        time.sleep(wait_interval)
        wait += wait_interval
        if wait >= max_wait:
            dataserver.stop()
            raise Exception("Server couldn't start up in time.")


def main():
    """Initialize and start the server."""
    import time
    import argparse

    from bcipy.acquisition.datastream.generator import file_data_generator
    from bcipy.acquisition.devices import preconfigured_device

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filename', default=None,
                        help="file containing data to be streamed; "
                        "if missing, random data will be served.")
    parser.add_argument('-m', '--markers', action="store_true", default=False)
    parser.add_argument('-n', '--name', default='DSI-24', help='Name of the device spec to mock.')
    args = parser.parse_args()

    if args.filename:
        daq_type, sample_rate, channels = _settings(args.filename)
        device_spec = preconfigured_device(daq_type)
        device_spec.sample_rate = sample_rate
        device_spec.channels = channels
        generator = file_data_generator(filename=args.filename)
    else:
        device_spec = preconfigured_device(args.name)
        generator = None

    markers = True if args.markers else False
    try:
        server = LslDataServer(device_spec=device_spec, generator=generator, add_markers=markers)

        log.debug("New server created")
        server.start()
        log.debug("Server started")
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Keyboard Interrupt")
        server.stop()


if __name__ == '__main__':
    # Run this as: python -m bcipy.acquisition.datastream.lsl_server
    main()



import os
import unittest
from unittest.mock import patch, mock_open

from mockito import any, mock, when, verify, unstub
import psychopy

from bcipy.helpers.exceptions import BciPyCoreException
from bcipy.helpers.triggers import (
    _calibration_trigger,
    FlushFrequency,
    TriggerType,
    Trigger,
    TriggerHandler
)


class TestCalibrationTrigger(unittest.TestCase):
    """Test Calibration Triggers.

    Unittests to assert the calibration trigger method. This is used
        during our tasks to reconcile timing between acquisition and displays.
    """

    clock = mock()
    display = mock()
    display.size = [500, 500]
    trigger_name = 'calibration_trigger'
    trigger_time = 1

    def setUp(self) -> None:
        unstub()

    def test_image_calibration_trigger(self):
        trigger_type = 'image'
        image_mock = mock()
        when(psychopy.visual).ImageStim(win=self.display,
                                        image=any(),
                                        pos=any(),
                                        mask=any(),
                                        ori=any()).thenReturn(image_mock)
        when(self.display).callOnFlip(any(), any(), any())
        when(image_mock).draw()
        when(self.display).flip()
        when(psychopy.core).wait(self.trigger_time)

        _calibration_trigger(self.clock, trigger_type, self.trigger_name,
                             self.trigger_time, self.display)

        verify(self.display, times=1).callOnFlip(any(), any(), any())
        verify(image_mock, times=1).draw()
        verify(self.display, times=1).flip()
        verify(psychopy.core, times=1).wait(self.trigger_time)

    def test_image_calibration_trigger_with_on_trigger(self):
        trigger_type = 'image'
        image_mock = mock()
        on_trigger = mock()
        when(psychopy.visual).ImageStim(win=self.display,
                                        image=any(),
                                        pos=any(),
                                        mask=any(),
                                        ori=any()).thenReturn(image_mock)
        when(self.display).callOnFlip(any(), any(), any())
        when(self.display).callOnFlip(on_trigger, self.trigger_name)
        when(image_mock).draw()
        when(self.display).flip()
        when(psychopy.core).wait(self.trigger_time)

        _calibration_trigger(self.clock, trigger_type, self.trigger_name,
                             self.trigger_time, self.display, on_trigger)

        verify(self.display, times=1).callOnFlip(any(), any(), any())
        verify(self.display, times=1).callOnFlip(on_trigger, self.trigger_name)
        verify(image_mock, times=1).draw()
        verify(self.display, times=1).flip()
        verify(psychopy.core, times=1).wait(self.trigger_time)

    def test_text_calibration_trigger(self):
        trigger_type = 'text'
        text_mock = mock()
        when(psychopy.visual).TextStim(self.display,
                                       text='').thenReturn(text_mock)
        when(self.display).callOnFlip(any(), any(), any())
        when(text_mock).draw()
        when(self.display).flip()
        when(psychopy.core).wait(self.trigger_time)

        _calibration_trigger(self.clock, trigger_type, self.trigger_name,
                             self.trigger_time, self.display)

        verify(self.display, times=1).callOnFlip(any(), any(), any())
        verify(text_mock, times=1).draw()
        verify(self.display, times=1).flip()
        verify(psychopy.core, times=1).wait(self.trigger_time)

    def test_text_calibration_trigger_with_on_trigger(self):
        trigger_type = 'text'
        text_mock = mock()
        on_trigger = mock()
        when(psychopy.visual).TextStim(self.display,
                                       text='').thenReturn(text_mock)
        when(self.display).callOnFlip(any(), any(), any())
        when(self.display).callOnFlip(on_trigger, self.trigger_name)
        when(text_mock).draw()
        when(self.display).flip()
        when(psychopy.core).wait(self.trigger_time)

        _calibration_trigger(self.clock, trigger_type, self.trigger_name,
                             self.trigger_time, self.display, on_trigger)

        verify(self.display, times=1).callOnFlip(any(), any(), any())
        verify(self.display, times=1).callOnFlip(on_trigger, self.trigger_name)
        verify(text_mock, times=1).draw()
        verify(self.display, times=1).flip()
        verify(psychopy.core, times=1).wait(self.trigger_time)

    def test_exception_invalid_calibration_trigger_type(self):
        trigger_type = 'invalid_type'
        with self.assertRaises(BciPyCoreException):
            _calibration_trigger(self.clock, trigger_type, self.trigger_name,
                                 self.trigger_time, self.display)

    def test_exception_no_display_calibration_trigger_type(self):
        trigger_type = 'image'
        with self.assertRaises(BciPyCoreException):
            _calibration_trigger(
                self.clock,
                trigger_type,
                self.trigger_name,
                self.trigger_time,
                None,
            )


class TestTrigger(unittest.TestCase):
    def setUp(self):
        self.label = 'A'
        self.type = TriggerType.NONTARGET
        self.time = 1
        self.test_trigger = Trigger(self.label,
                                    self.type,
                                    self.time)

    def test_create_trigger(self):
        self.assertTrue(self.test_trigger.label == self.label and
                        self.test_trigger.type == self.type and
                        self.test_trigger.time == self.time)

    def test_print_trigger(self):
        expected = f'Trigger: label=[{self.label}] type=[{self.type}] time=[{self.time}]'
        result = self.test_trigger.__repr__()
        self.assertEqual(expected, result)

    def test_from_list(self):
        trg = Trigger('A', TriggerType.NONTARGET, 1)
        self.assertEqual(Trigger.from_list(['A', 'nontarget', '1']), trg)

    def test_with_offset(self):
        trg = Trigger('A', TriggerType.NONTARGET, 1)
        self.assertEqual(trg.with_offset(1).time, 2)


class TestTriggerHandler(unittest.TestCase):

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def setUp(self, mock_file):
        self.mock_file = mock_file
        self.path_name = '.'
        self.file_name = 'test'
        self.flush = FlushFrequency.END
        self.file = f'{self.path_name}/{self.file_name}.txt'
        # with patch('builtins.open', mock_open(read_data='data')) as _:
        self.handler = TriggerHandler(self.path_name, self.file_name, self.flush)
        self.mock_file.assert_called_once_with(self.file, 'w+', encoding=self.handler.encoding)

    def tearDown(self):
        unstub()

    def test_file_exist_exception(self):
        with open(self.file, 'w+', encoding=self.handler.encoding) as _:
            with self.assertRaises(Exception):
                TriggerHandler(self.path_name, self.file_name, FlushFrequency.END)
        os.remove(self.file)

    def test_add_triggers_returns_list_of_triggers(self):
        trigger = Trigger('A', TriggerType.NONTARGET, 1)
        inquiry_triggers = [trigger]

        response = self.handler.add_triggers(inquiry_triggers)
        self.assertEqual(response, inquiry_triggers)

    def test_write_triggers_flushes_triggers(self):
        inquiry_triggers = [Trigger('A', TriggerType.NONTARGET, 1)]
        self.handler.add_triggers(inquiry_triggers)
        self.assertNotEqual(self.handler.triggers, [])
        self.handler.write()
        self.assertEqual(self.handler.triggers, [])

    def test_add_triggers_calls_write_when_flush_every(self):
        self.handler.flush = FlushFrequency.EVERY
        when(self.handler).write().thenReturn()
        inquiry_triggers = [Trigger('A', TriggerType.NONTARGET, 1)]
        self.handler.add_triggers(inquiry_triggers)

        verify(self.handler, times=1).write()

    def test_load_returns_list_of_triggers(self):
        trg = Trigger('A', TriggerType.NONTARGET, 1)
        when(TriggerHandler).read_text_file(any()).thenReturn(([trg], 0.0))

        response = self.handler.load('test_path_not_real')
        self.assertEqual(response[0], trg)

    def test_load_applies_offset(self):
        trg = Trigger('A', TriggerType.NONTARGET, 1)

        when(TriggerHandler).read_text_file(any()).thenReturn(([trg], 0.0))
        response = self.handler.load('test_path_not_real', offset=1)
        self.assertEqual(response[0].time, 2)

    def test_load_exclusion(self):
        fixation_trg = Trigger('+', TriggerType.FIXATION, 2)
        trg_list = [Trigger('A', TriggerType.NONTARGET, 1), fixation_trg]

        when(TriggerHandler).read_text_file(any()).thenReturn((trg_list, 0.0))

        response = self.handler.load('test_path_not_real',
                                     exclusion=[TriggerType.NONTARGET])
        self.assertEqual(response[0], fixation_trg)

    @patch('bcipy.helpers.triggers.os.path.exists')
    def test_read_data(self, path_exists_mock):
        """Test that trigger data is correctly read."""
        trg_data = '''starting_offset offset 3.47
                    J prompt 6.15
                    + fixation 8.11
                    F nontarget 8.58
                    D nontarget 8.88
                    J target 9.18
                    T nontarget 9.49
                    K nontarget 9.79
                    _ nontarget 11.30'''
        path_exists_mock.returnValue = True
        with patch('builtins.open', mock_open(read_data=trg_data),
                   create=True):
            triggers, offset = TriggerHandler.read_text_file('triggers.txt')
            self.assertEqual(len(triggers), 9)
            self.assertEqual(triggers[1].label, 'J')
            self.assertEqual(triggers[1].type, TriggerType.PROMPT)
            self.assertEqual(triggers[1].time, 6.15)
            self.assertEqual(offset, 3.47)

    @patch('bcipy.helpers.triggers.os.path.exists')
    def test_read_data_bad_format(self, path_exists_mock):
        """Test that exception is thrown when trigger type doesn't exist."""
        trg_data = '''start_offset offset
                    + fixation 8.11
                    F hello_world 8.58
                    system_data system 6.23'''
        path_exists_mock.returnValue = True
        with patch('builtins.open', mock_open(read_data=trg_data),
                   create=True):
            with self.assertRaises(BciPyCoreException) as ctx:
                _trg, _ = TriggerHandler.read_text_file('triggers.txt')

            self.assertIn("line 1", ctx.exception.message)

    @patch('bcipy.helpers.triggers.os.path.exists')
    def test_read_data_bad_trigger_type(self, path_exists_mock):
        """Test that exception is thrown when trigger type doesn't exist."""
        trg_data = '''start_offset offset 3.47
                    + fixation 8.11
                    F hello_world 8.58
                    system_data system 6.23'''
        path_exists_mock.returnValue = True
        with patch('builtins.open', mock_open(read_data=trg_data),
                   create=True):
            with self.assertRaises(BciPyCoreException) as ctx:
                _trg, _ = TriggerHandler.read_text_file('triggers.txt')

            self.assertIn("line 3", ctx.exception.message)

    @patch('bcipy.helpers.triggers.os.path.exists')
    def test_read_data_bad_timestamp(self, path_exists_mock):
        """Test that exception is thrown when timestamp can't be converted."""
        trg_data = '''starting_offset offset 3.47
                    J prompt 6.15abc
                    + fixation 8.11
                    F hello_world 8.58
                    offset offset_correction 6.23'''
        path_exists_mock.returnValue = True
        with patch('builtins.open', mock_open(read_data=trg_data),
                   create=True):
            with self.assertRaises(BciPyCoreException) as ctx:
                _trg, _ = TriggerHandler.read_text_file('triggers.txt')

            self.assertIn("line 2", ctx.exception.message)

    @patch('bcipy.helpers.triggers.os.path.exists')
    def test_read_data_unicode(self, path_exists_mock):
        """Test that trigger data is correctly read."""
        trg_data = '''starting_offset offset 3.47
                    ■ prompt 6.15
                    + fixation 8.11
                    F nontarget 8.58
                    D nontarget 8.88
                    ■ target 9.18
                    T nontarget 9.49
                    K nontarget 9.79
                    _ nontarget 11.30'''
        path_exists_mock.returnValue = True
        with patch('builtins.open', mock_open(read_data=trg_data),
                   create=True):
            triggers, offset = TriggerHandler.read_text_file('triggers.txt')
            self.assertEqual(len(triggers), 9)
            self.assertEqual(triggers[1].label, '■')
            self.assertEqual(triggers[1].type, TriggerType.PROMPT)
            self.assertEqual(triggers[1].time, 6.15)
            self.assertEqual(offset, 3.47)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List
import numpy as np
from bcipy.helpers.stimuli import Reshaper


class SignalModel(ABC):
    @property
    @abstractmethod
    def reshaper(self) -> Reshaper:
        """Reshapes data into trials or inquiry as needed for each model."""
        ...

    @abstractmethod
    def fit(self, training_data: np.ndarray, training_labels: np.ndarray):
        """
        Train the model using the provided data and labels.
        Return self for convenience.
        """
        ...

    @abstractmethod
    def evaluate(self, test_data: np.ndarray, test_labels: np.ndarray):
        """Compute model performance characteristics on the provided test data and labels."""
        ...

    @abstractmethod
    def predict(self, data: np.ndarray, inquiry: List[str], symbol_set: List[str]) -> np.ndarray:
        """
        Using the provided data, compute log likelihoods over the entire symbol set.
        Args:
            inquiry - the subset of symbols presented
            symbol_set - the entire alphabet of symbols
        """
        ...

    @abstractmethod
    def save(self, path: Path) -> None:
        """Save model state to the provided checkpoint"""
        ...

    @abstractmethod
    def load(self, path: Path) -> None:
        """Restore model state from the provided checkpoint"""
        ...


class ModelEvaluationReport:
    """
    Describes model performance characteristics.
    """

    def __init__(self, auc: float):
        self.auc = auc

    def __eq__(self, other):
        return np.allclose(self.auc, other.auc)



"""Module for functionality related to system configuration"""
from codecs import open as codecsopen
from collections import abc
from json import dump, load
from pathlib import Path
from typing import Any, Dict, NamedTuple, Tuple

from bcipy.config import DEFAULT_ENCODING, DEFAULT_PARAMETERS_PATH


class Parameter(NamedTuple):
    """Represents a single parameter"""
    value: Any
    section: str
    readableName: str
    helpTip: str
    recommended_values: list
    type: str


class ParameterChange(NamedTuple):
    """Represents a Parameter that has been modified from a different value."""
    parameter: Parameter
    original_value: Any


class Parameters(dict):
    """Configuration parameters for BciPy.

        source: str - optional path to a JSON file. If file exists, data will be
            loaded from here. Raises an exception unless the entries are dicts with
            the required_keys.

        cast_values: bool - if True cast values to specified type; default is False.
        """

    def __init__(self, source: str = None, cast_values: bool = False):
        super().__init__()
        self.source = source
        self.cast_values = cast_values

        self.required_keys = set([
            'value', 'section', 'readableName', 'helpTip',
            'recommended_values', 'type'
        ])
        self.conversions = {
            'int': int,
            'float': float,
            'bool': lambda val: val == 'true' or val is True,
            'str': str,
            'directorypath': str,
            'filepath': str
        }
        self.load_from_source()

    @classmethod
    def from_cast_values(cls, **kwargs):
        """Create a new Parameters object from cast values. This is useful
        primarily for testing

        >>> Parameters.from_cast_values(time_prompt=1.0, fake_data=True)
        """
        params = Parameters(source=None, cast_values=True)
        for key, val in kwargs.items():
            value_type = type(val).__name__
            value_str = str(val).lower() if value_type == 'bool' else str(val)
            params.add_entry(
                key, {
                    'value': value_str,
                    'section': '',
                    'readableName': '',
                    'helpTip': '',
                    'recommended_values': '',
                    'type': value_type
                })
        return params

    @property
    def supported_types(self):
        """Supported types for casting values"""
        return self.conversions.keys()

    def cast_value(self, entry: dict):
        """Takes an entry with a desired type and attempts to cast it to that type."""
        cast = self.conversions[entry['type']]
        return cast(entry['value'])

    def serialized_value(self, value, entry_type):
        """Convert a value back into its serialized form"""
        serialized = str(value)
        return serialized.lower() if entry_type == 'bool' else serialized

    def __getitem__(self, key):
        """Override to handle cast values"""
        entry = self.get_entry(key)
        if self.cast_values:
            return self.cast_value(entry)
        return entry

    def __setitem__(self, key, value):
        """Override to handle cast values"""
        if self.cast_values:
            # Can only set values for existing entries when cast.
            entry = self.get_entry(key)
            entry['value'] = self.serialized_value(value, entry['type'])
        else:
            self.add_entry(key, value)

    def add_entry(self, key, value):
        """Adds a configuration parameter."""
        self.check_valid_entry(key, value)
        super().__setitem__(key, value)

    def get_entry(self, key):
        """Get the non-cast entry associated with the given key."""
        return super().__getitem__(key)

    def get(self, key, d=None):
        """Override to handle cast values"""
        entry = super().get(key, d)
        if self.cast_values and entry != d:
            return self.cast_value(entry)
        return entry

    def entries(self):
        """Uncast items"""
        return super().items()

    def items(self):
        """Override to handle cast values"""
        if self.cast_values:
            return [(key, self.cast_value(entry))
                    for key, entry in self.entries()]
        return self.entries()

    def values(self):
        """Override to handle cast values"""
        vals = super().values()
        if self.cast_values:
            return [self.cast_value(entry) for entry in vals]
        return vals

    def update(self, *args, **kwargs):
        """Override to ensure update uses __setitem___"""
        for key, value in dict(*args, **kwargs).items():
            self[key] = value

    def copy(self):
        """Override
        """
        params = Parameters(source=None, cast_values=self.cast_values)
        params.load(super().copy())
        return params

    def load(self, data: dict):
        """Load values from a dict, validating entries (see check_valid_entry) and raising
        an exception for invalid values.

        data: dict of configuration parameters.
        """
        for name, entry in data.items():
            self.add_entry(name, entry)

    def load_from_source(self):
        """Load data from the configured JSON file."""
        if self.source:
            with codecsopen(self.source, 'r',
                            encoding=DEFAULT_ENCODING) as json_file:
                data = load(json_file)
                self.load(data)

    def check_valid_entry(self, entry_name: str, entry: dict):
        """Checks if the given entry is valid. Raises an exception unless the entry is formatted:

        "fake_data": {
            "value": "true",
            "section": "bci_config",
            "readableName": "Fake Data Sessions",
            "helpTip": "If true, fake data server used",
            "recommended_values": "",
            "type": "bool"
        }

        entry_name : str - name of the configuration parameter
        entry : dict - parameter properties
        """
        if not isinstance(entry, abc.Mapping):
            raise AttributeError(f"'{entry_name}' value must be a dict")
        if set(entry.keys()) != self.required_keys:
            raise Exception(
                f"Incorrect format for key: {entry_name}; value must contain required keys"
            )
        if entry['type'] not in self.supported_types:
            raise Exception(
                f"Type not supported for key: {entry_name}, type: {entry['type']}"
            )
        if entry['type'] == "bool" and entry['value'] not in ['true', 'false']:
            raise Exception(
                f"Invalid value for key: {entry_name}. Must be either 'true' or 'false'"
            )

    def source_location(self) -> Tuple[Path, str]:
        """Location of the source json data if source was provided.

        Returns Tuple(Path, filename: str)
        """
        if self.source:
            path = Path(self.source)
            return (path.parent, path.name)
        return (None, None)

    def save(self, directory: str = None, name: str = None):
        """Save parameters to the given location

        directory: str - optional location to save; default is the source_directory.
        name: str - optional name of new parameters file; default is the source filename.

        Returns the path of the saved file.
        """
        if (not directory or not name) and not self.source:
            raise AttributeError('name and directory parameters are required')

        source_directory, source_name = self.source_location()
        location = directory if directory else source_directory
        filename = name if name else source_name
        path = Path(location, filename)
        with open(path, 'w', encoding=DEFAULT_ENCODING) as json_file:
            dump(dict(self.entries()), json_file, ensure_ascii=False, indent=2)
        return str(path)

    def add_missing_items(self, parameters) -> bool:
        """Given another Parameters instance, add any items that are not already
        present. Existing items will not be updated.

        parameters: Parameters - object from which to add parameters.

        Returns bool indicating whether or not any new items were added.
        """
        updated = False
        existing_keys = self.keys()
        for key, val in parameters.entries():
            if key not in existing_keys:
                self.add_entry(key, val)
                updated = True
        return updated

    def diff(self, parameters) -> Dict[str, ParameterChange]:
        """Lists the differences between this and another set of parameters.
        A None original_value indicates a new parameter.

        Parameters
        ----------
            parameters : Parameters - set of parameters for comparison; these
                are considered the original values and the current set the
                changed values.
        """
        diffs = {}

        for key, param in self.entries():
            if key in parameters.keys():
                original = parameters.get_entry(key)
                if self.cast_value(original) != self.cast_value(param):
                    diffs[key] = ParameterChange(
                        parameter=param, original_value=original['value'])
            else:
                diffs[key] = ParameterChange(parameter=param,
                                             original_value=None)
        return diffs


def changes_from_default(source: str) -> Dict[str, ParameterChange]:
    """Determines which parameters have changed from the default params.

    Parameters
    ----------
        source - path to the parameters json file that will be compared with
            the default parameters.
    """
    default = Parameters(source=DEFAULT_PARAMETERS_PATH, cast_values=True)
    params = Parameters(source=source, cast_values=True)
    return params.diff(default)



"""Defines the language model base class."""
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional, Tuple
import json

from bcipy.helpers.exceptions import UnsupportedResponseType
from bcipy.helpers.symbols import DEFAULT_SYMBOL_SET
from bcipy.config import DEFAULT_LM_PARAMETERS_PATH


class ResponseType(Enum):
    """Language model response type options."""
    SYMBOL = 'Symbol'
    WORD = 'Word'

    def __str__(self):
        return self.value


class LanguageModel(ABC):
    """Parent class for Language Models."""

    _response_type: ResponseType = None
    symbol_set: List[str] = None

    def __init__(self,
                 response_type: Optional[ResponseType] = None,
                 symbol_set: Optional[List[str]] = None):
        self.response_type = response_type or ResponseType.SYMBOL
        self.symbol_set = symbol_set or DEFAULT_SYMBOL_SET
        with open(DEFAULT_LM_PARAMETERS_PATH, 'r') as params_file:
            self.parameters = json.load(params_file)

    @classmethod
    def name(cls) -> str:
        """Model name used for configuration"""
        suffix = 'LanguageModel'
        if cls.__name__.endswith(suffix):
            return cls.__name__[0:-len(suffix)].upper()
        return cls.__name__.upper()

    @abstractmethod
    def supported_response_types(self) -> List[ResponseType]:
        """Returns a list of response types supported by this language model."""

    @property
    def response_type(self) -> ResponseType:
        """Returns the current response type"""
        return self._response_type

    @response_type.setter
    def response_type(self, value: ResponseType):
        """Attempts to set the response type to the given value"""
        if value not in self.supported_response_types():
            raise UnsupportedResponseType(
                f"{value} responses are not supported by this model")
        self._response_type = value

    @abstractmethod
    def predict(self, evidence: List[str]) -> List[Tuple]:
        """
        Using the provided data, compute log likelihoods over the entire symbol set.
        Args:
            evidence - ['A', 'B']

        Response:
            probability - dependant on response type, a list of words or symbols with probability
        """
        ...

    @abstractmethod
    def update(self) -> None:
        """Update the model state"""
        ...

    @abstractmethod
    def load(self) -> None:
        """Restore model state from the provided checkpoint"""
        ...

    def reset(self) -> None:
        """Reset language model state"""
        ...

    def state_update(self, evidence: List[str]) -> List[Tuple]:
        """Update state by predicting and updating"""



from abc import ABC, abstractmethod
from logging import Logger
from typing import List, Tuple, Union

from psychopy import visual

from bcipy.helpers.clock import Clock
from bcipy.helpers.system_utils import get_screen_info


class Display(ABC):
    """Display.

    Base class for BciPy displays. This defines the logic necessary for task executions that require a display.
    """

    window: visual.Window = None
    timing_clock: Clock = None
    experiment_clock: Clock = None
    logger: Logger = None
    stimuli_inquiry: List[str] = None
    stimuli_colors: List[str] = None
    stimuli_timing: List[float] = None
    task = None

    @abstractmethod
    def do_inquiry(self) -> List[float]:
        """Do inquiry.

        Animates an inquiry of stimuli and returns a list of stimuli trigger timing.
        """
        ...

    @abstractmethod
    def wait_screen(self) -> None:
        """Wait Screen.

        Define what happens on the screen when a user pauses a session.
        """
        ...

    @abstractmethod
    def update_task_bar(self, *args, **kwargs) -> None:
        """Update Task.

        Update any taskbar-related display items not related to the inquiry. Ex. stimuli count 1/200.
        """
        ...

    def schedule_to(self, stimuli: list, timing: list, colors: list) -> None:
        """Schedule To.

        Schedule stimuli elements (works as a buffer) before calling do_inquiry.
        """
        ...

    def draw_static(self) -> None:
        """Draw Static.

        Displays task information not related to the inquiry.
        """
        ...

    def preview_inquiry(self) -> List[float]:
        """Preview Inquiry.

        Display an inquiry or instruction beforehand to the user. This should be called before do_inquiry.
        This can be used to determine if the desired stimuli is present before displaying them more laboriusly
        or prompting users before the inquiry.
        All stimuli elements (stimuli, timing, colors) must be set on the display before calling this method.
        This implies, something like schedule_to is called.
        """
        ...


def init_display_window(parameters):
    """
    Init Display Window.

    Function to Initialize main display window
        needed for all later stimuli presentation.

    See Psychopy official documentation for more information and working demos:
        http://www.psychopy.org/api/visual/window.html
    """

    # Check is full_screen mode is set and get necessary values
    if parameters['full_screen']:

        # set window attributes based on resolution
        screen_info = get_screen_info()
        window_height = screen_info.height
        window_width = screen_info.width

        # set full screen mode to true (removes os dock, explorer etc.)
        full_screen = True

    # otherwise, get user defined window attributes
    else:

        # set window attributes directly from parameters file
        window_height = parameters['window_height']
        window_width = parameters['window_width']

        # make sure full screen is set to false
        full_screen = False

    # Initialize PsychoPy Window for Main Display of Stimuli
    display_window = visual.Window(
        size=[window_width,
              window_height],
        screen=parameters['stim_screen'],
        allowGUI=False,
        useFBO=False,
        fullscr=full_screen,
        allowStencil=False,
        monitor='mainMonitor',
        winType='pyglet', units='norm', waitBlanking=False,
        color=parameters['background_color'])

    # Return display window to caller
    return display_window


class StimuliProperties:
    """"Stimuli Properties.

    An encapsulation of properties relevant to core stimuli presentation in an RSVP or Matrix paradigm.
    """

    def __init__(
            self,
            stim_font: str,
            stim_pos: Tuple[float, float],
            stim_height: float,
            stim_inquiry: List[str] = None,
            stim_colors: List[str] = None,
            stim_timing: List[float] = None,
            is_txt_stim: bool = True,
            prompt_time: float = None):
        """Initialize Stimuli Parameters.

        stim_font(List[str]): Ordered list of colors to apply to information stimuli
        stim_pos(Tuple[float, float]): Position on window where the stimuli will be presented
        stim_height(float): Height of all stimuli
        stim_inquiry(List[str]): Ordered list of text to build stimuli with
        stim_colors(List[str]): Ordered list of colors to apply to stimuli
        stim_timing(List[float]): Ordered list of timing to apply to an inquiry using the stimuli
        is_txt_stim(bool): Whether or not this is a text based stimuli (False implies image based)
        prompt_time(float): Time to display target prompt for at the beginning of inquiry
        """
        self.stim_font = stim_font
        self.stim_pos = stim_pos
        self.stim_height = stim_height
        self.stim_inquiry = stim_inquiry or []
        self.stim_colors = stim_colors or []
        self.stim_timing = stim_timing or []
        self.is_txt_stim = is_txt_stim
        self.stim_length = len(self.stim_inquiry)
        self.sti = None
        self.prompt_time = prompt_time

    def build_init_stimuli(self, window: visual.Window) -> Union[visual.TextStim, visual.ImageStim]:
        """"Build Initial Stimuli.

        This method constructs the stimuli object which can be updated later. This is more
            performant than creating a new stimuli each call. It can create either an image or text stimuli
            based on the boolean self.is_txt_stim.
        """
        if self.is_txt_stim:
            self.sti = visual.TextStim(
                win=window,
                color='white',
                height=self.stim_height,
                text='',
                font=self.stim_font,
                pos=self.stim_pos,
                wrapWidth=None, colorSpace='rgb',
                opacity=1, depth=-6.0)
        else:
            self.sti = visual.ImageStim(
                win=window,
                image=None,
                mask=None,
                pos=self.stim_pos,
                ori=0.0)
        return self.sti


class InformationProperties:
    """"Information Properties.

    An encapsulation of properties relevant to task information presentation in an RSVP paradigm. This could be
        messaging relevant to feedback or static text to remain on screen not related to task tracking.
    """

    def __init__(
            self,
            info_color: List[str],
            info_text: List[str],
            info_font: List[str],
            info_pos: List[Tuple[float, float]],
            info_height: List[float]):
        """Initialize Information Parameters.

        info_color(List[str]): Ordered list of colors to apply to information stimuli
        info_text(List[str]): Ordered list of text to apply to information stimuli
        info_font(List[str]): Ordered list of font to apply to information stimuli
        info_pos(Tuple[float, float]): Position on window where the Information stimuli will be presented
        info_height(List[float]): Ordered list of height of Information stimuli
        """
        self.info_color = info_color
        self.info_text = info_text
        self.info_font = info_font
        self.info_pos = info_pos
        self.info_height = info_height

    def build_info_text(self, window: visual.Window) -> List[visual.TextStim]:
        """"Build Information Text.

        Constructs a list of Information stimuli to display.
        """
        self.text_stim = []
        for idx in range(len(self.info_text)):
            self.text_stim.append(visual.TextStim(
                win=window,
                color=self.info_color[idx],
                height=self.info_height[idx],
                text=self.info_text[idx],
                font=self.info_font[idx],
                pos=self.info_pos[idx],
                wrapWidth=None, colorSpace='rgb',
                opacity=1, depth=-6.0))
        return self.text_stim


class PreviewInquiryProperties:
    """"Preview Inquiry Properties.
    An encapsulation of properties relevant to preview_inquiry() operation.
    """

    def __init__(
            self,
            preview_only: bool,
            preview_inquiry_length: float,
            preview_inquiry_progress_method: int,
            preview_inquiry_key_input: str,
            preview_inquiry_isi: float):
        """Initialize Inquiry Preview Parameters.
        preview_only(bool): If True, only preview the inquiry and do not probe for response
        preview_inquiry_length(float): Length of time in seconds to present the inquiry preview
        preview_inquiry_progress_method(int): Method of progression for inquiry preview.
            0 == preview only; 1 == press to accept inquiry; 2 == press to skip inquiry.
        preview_inquiry_key_input(str): Defines which key should be listened to for progressing
        preview_inquiry_isi(float): Length of time after displaying the inquiry preview to display a blank screen
        """
        self.preview_inquiry_length = preview_inquiry_length
        self.preview_inquiry_key_input = preview_inquiry_key_input
        self.press_to_accept = True if preview_inquiry_progress_method == 1 else False
        self.preview_only = preview_only
        self.preview_inquiry_isi = preview_inquiry_isi


class VEPStimuliProperties(StimuliProperties):

    def __init__(
            self,
            stim_font: str,
            stim_pos: List[Tuple[float, float]],
            stim_height: float,
            timing: Tuple[float, float, float] = None,
            stim_color: List[List[str]] = None,
            inquiry: List[List[str]] = None,
            stim_length: int = 1):
        """Initialize VEP Stimuli Parameters.
        stim_color(List[str]): Ordered list of colors to apply to VEP stimuli
        stim_font(str): Font to apply to all VEP stimuli
        stim_pos(List[Tuple[float, float]]): Position on the screen where to present to VEP text
        stim_height(float): Height of all VEP text stimuli
        """
        # static properties
        self.stim_font = stim_font
        self.stim_height = stim_height
        self.is_txt_stim = True
        self.stim_length = stim_length  # how many times to flicker
        self.stim_pos = stim_pos

        # dynamic property. List of length 3. 1. prompt; 2. fixation; 3. inquiry
        self.stim_timing = timing

        # dynamic properties, must be a a list of lists where each list is a different box
        self.stim_colors = stim_color
        self.stim_inquiry = inquiry

    def build_init_stimuli(self, window: visual.Window) -> None:
        """"Build Initial Stimuli."""
        ...



"""Functionality for managing multiple devices."""
import logging
from typing import Dict, List, Optional
from bcipy.acquisition.protocols.lsl.lsl_client import LslAcquisitionClient
from bcipy.acquisition.devices import DeviceSpec

log = logging.getLogger(__name__)


class ClientManager():
    """Manages multiple acquisition clients. This class can also act as an
    acquisition client. If used in this way, it dispatches to the managed
    client with the default_client_type.

    >>> from bcipy.acquisition import LslAcquisitionClient
    >>> from bcipy.acquisition.devices import DeviceSpec
    >>> spec = DeviceSpec('Testing', ['ch1', 'ch2', 'ch3'], 60.0, 'EEG')
    >>> manager = ClientManager()
    >>> eeg_client = LslAcquisitionClient(device_spec=spec)
    >>> manager.add_client(eeg_client)
    >>> manager.device_spec == spec
    True

    Parameters
    ----------
        default_content_type - used for dispatching calls to an LslClient.
    """

    def __init__(self, default_content_type: str = 'EEG'):
        self._clients: Dict[str, LslAcquisitionClient] = {}
        self.default_content_type = default_content_type

    @property
    def clients(self) -> List[LslAcquisitionClient]:
        """Returns a list of the managed clients."""
        return self._clients.values()

    @property
    def device_specs(self) -> List[DeviceSpec]:
        """Returns a list of DeviceSpecs for all the clients."""
        return [client.device_spec for client in self.clients]

    @property
    def default_client(self) -> Optional[LslAcquisitionClient]:
        """Returns the default client."""
        return self.get_client(self.default_content_type)

    def add_client(self, client: LslAcquisitionClient):
        """Add the given client to the manager."""
        self._clients[client.device_spec.content_type] = client

    def get_client(self, content_type: str) -> Optional[LslAcquisitionClient]:
        """Get client by content type"""
        return self._clients.get(content_type, None)

    def start_acquisition(self):
        """Start acquiring data for all clients"""
        for client in self.clients:
            log.info(f"Connecting to {client.device_spec.name}...")
            client.start_acquisition()

    def stop_acquisition(self):
        """Stop acquiring data for all clients"""
        for client in self.clients:
            client.stop_acquisition()

    def cleanup(self):
        """Perform any cleanup tasks"""
        for client in self.clients:
            client.cleanup()

    def __getattr__(self, name):
        """Dispatch unknown properties and methods to the client with the
        default content type."""
        client = self.default_client
        if client:
            return client.__getattribute__(name)
        raise AttributeError(f"Missing attribute: {name}")



"""Config.

This module contains the configuration for the BciPy application.
This includes the default parameters, static paths and core experiment configuration.
"""
from pathlib import Path

DEFAULT_ENCODING = 'utf-8'

# experiment configuration
DEFAULT_EXPERIMENT_ID = 'default'
EXPERIMENT_FILENAME = 'experiments.json'
FIELD_FILENAME = 'fields.json'
EXPERIMENT_DATA_FILENAME = 'experiment_data.json'
BCIPY_ROOT = Path(__file__).resolve().parent
ROOT = BCIPY_ROOT.parent
DEFAULT_EXPERIMENT_PATH = f'{ROOT}/.bcipy/experiment'
DEFAULT_FIELD_PATH = f'{ROOT}/.bcipy/field'

DEFAULT_PARAMETER_FILENAME = 'parameters.json'
DEFAULT_PARAMETERS_PATH = f'{BCIPY_ROOT}/parameters/{DEFAULT_PARAMETER_FILENAME}'
DEFAULT_DEVICE_SPEC_FILENAME = 'devices.json'
DEVICE_SPEC_PATH = f'{BCIPY_ROOT}/parameters/{DEFAULT_DEVICE_SPEC_FILENAME}'
DEFAULT_LM_PARAMETERS_FILENAME = 'lm_params.json'
DEFAULT_LM_PARAMETERS_PATH = f'{BCIPY_ROOT}/parameters/{DEFAULT_LM_PARAMETERS_FILENAME}'

STATIC_PATH = f'{BCIPY_ROOT}/static'
STATIC_IMAGES_PATH = f'{STATIC_PATH}/images'
STATIC_AUDIO_PATH = f'{STATIC_PATH}/sounds'
BCIPY_LOGO_PATH = f'{STATIC_IMAGES_PATH}/gui/cambi.png'
PREFERENCES_PATH = f'{ROOT}/.bcipy/bcipy_cache'
LM_PATH = f'{BCIPY_ROOT}/language/lms'

# core data configuration
RAW_DATA_FILENAME = 'raw_data'
TRIGGER_FILENAME = 'triggers.txt'
SESSION_DATA_FILENAME = 'session.json'
SESSION_SUMMARY_FILENAME = 'session.xlsx'
LOG_FILENAME = 'bcipy_system_log.txt'

# misc configuration
WAIT_SCREEN_MESSAGE = 'Press Space to start or Esc to exit'
SESSION_COMPLETE_MESSAGE = 'Complete! Saving data...'
REMOTE_SERVER = "https://github.com/CAMBI-tech/BciPy/"



"""Utility functions for list processing."""
from typing import Callable, List
from itertools import zip_longest


def destutter(items: List, key: Callable = lambda x: x) -> List:
    """Removes sequential duplicates from a list. Retains the last item in the
    sequence. Equality is determined using the provided key function.

    Parameters
    ----------
        items - list of items with sequential duplicates
        key - equality function
    """
    deduped = []
    for item in items:
        if len(deduped) == 0 or key(item) != key(deduped[-1]):
            deduped.append(item)
        else:
            deduped[-1] = item
    return deduped


def grouper(iterable, chunk_size, incomplete="fill", fillvalue=None):
    "Collect data into non-overlapping fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, fillvalue='x') --> ABC DEF Gxx
    # grouper('ABCDEFG', 3, incomplete='ignore') --> ABC DEF
    chunks = [iter(iterable)] * chunk_size
    if incomplete == "fill":
        if fillvalue:
            return zip_longest(*chunks, fillvalue=fillvalue)
        raise ValueError('fillvalue must be defined if using incomplete=fill')
    if incomplete == "ignore":
        return zip(*chunks)

    raise ValueError("Expected fill or ignore")



"""Tests for session helpers."""

import filecmp
import os
import shutil
import tempfile
import unittest
from pathlib import Path

from bcipy.config import SESSION_DATA_FILENAME
from bcipy.helpers.session import read_session, session_csv, session_data


class TestSessionHelper(unittest.TestCase):
    """Tests for session helper."""

    def setUp(self):
        """Override; set up the needed path for load functions."""
        self.data_dir = f"{os.path.dirname(__file__)}/resources/mock_session/"
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Override"""
        shutil.rmtree(self.temp_dir)

    def test_session_data(self):
        """Test session_data method outputs letters with evidence."""
        data = session_data(data_dir=self.data_dir)
        self.assertEqual(data["total_number_series"], 2)
        self.assertTrue("A" in data["series"]["1"]["0"]["likelihood"])
        self.assertTrue("B" in data["series"]["1"]["0"]["likelihood"])

    def test_read_session(self):
        """Test reading data from json."""
        session = read_session(Path(self.data_dir, SESSION_DATA_FILENAME))
        self.assertEqual(session.total_number_series, 2)

    def test_session_csv(self):
        """Test functionality to transform session.json file to a csv
        summarizing the evidence."""

        session = read_session(Path(self.data_dir, SESSION_DATA_FILENAME))
        csv_name = str(Path(self.temp_dir, 'mock_session.csv'))

        session_csv(session, csv_file=csv_name)

        generated = Path(self.temp_dir, 'mock_session.csv')
        expected = Path(self.data_dir, "session.csv")

        self.assertTrue(generated.is_file())
        self.assertTrue(filecmp.cmp(expected, generated))



"""Functionality for converting the bcipy raw data output to other formats"""
import logging
import os
import tarfile
from pathlib import Path
from typing import Dict, List, Tuple, Optional

import numpy as np

from pyedflib import FILETYPE_EDFPLUS, EdfWriter, FILETYPE_BDFPLUS
from tqdm import tqdm

from bcipy.config import RAW_DATA_FILENAME, TRIGGER_FILENAME, DEFAULT_PARAMETER_FILENAME
from bcipy.helpers.load import load_json_parameters, load_raw_data
from bcipy.helpers.raw_data import RawData
from bcipy.signal.process import Composition, get_default_transform
from bcipy.helpers.triggers import trigger_decoder, trigger_durations

import mne
from mne.io import RawArray


logger = logging.getLogger(__name__)

FILE_LENGTH_LIMIT = 150


def convert_to_edf(data_dir: str,
                   edf_path: str = None,
                   overwrite: bool = False,
                   write_targetness: bool = False,
                   use_event_durations: bool = False,
                   remove_pre_fixation: bool = False,
                   pre_filter: bool = False) -> Path:
    """ Converts BciPy data to the EDF+ filetype using pyEDFlib.

    See https://www.edfplus.info/ for more detailed information about the edf specification.
    See https://www.teuniz.net/edfbrowser/ for a free EDF/BDF viewer.

    Parameters
    ----------
    data_dir - directory which contains the data to be converted. This location must contain:
        1. a parameter file.
        2. a raw_data.csv file.
        3. a trigger.txt file.
    edf_path - optional; path to write converted data; defaults to writing a file named raw_data.edf in the data_dir.
        Must end in .edf.
    overwrite - If True, the destination file (if it exists) will be overwritten. If False (default), an error will
        be raised if the file exists.
    write_targetness - If True, and targetness information is available, write that instead of the stimuli markers.
        False by default.
    use_event_durations - optional; if True assigns a duration to each event.
    remove_pre_fixation - optional; if True removes the non-inquiry based markers from the trigger data.
    pre_filter - optional; if True, apply the default filter to the data using to loaded parameters file.

    Returns
    -------
        Path to new edf file
    """
    if not edf_path:
        edf_path = str(Path(data_dir, f'{RAW_DATA_FILENAME}.edf').resolve())
    if not edf_path.endswith('.edf'):
        raise ValueError(f'edf_path=[{edf_path}] must end in .edf')

    data, channels, fs, events, annotation_channels, pre_filter = pyedf_convert(
        data_dir,
        write_targetness=write_targetness,
        use_event_durations=use_event_durations,
        remove_pre_fixation=remove_pre_fixation,
        pre_filter=pre_filter)

    return write_pyedf(
        edf_path,
        data,
        channels,
        fs,
        events,
        FILETYPE_EDFPLUS,
        overwrite,
        annotation_channels,
        pre_filter)


def convert_to_bdf(data_dir: str,
                   bdf_path: str = None,
                   overwrite: bool = False,
                   write_targetness: bool = False,
                   use_event_durations: bool = False,
                   remove_pre_fixation: bool = False,
                   pre_filter: bool = False) -> Path:
    """ Converts BciPy data to the BDF+ filetype using pyEDFlib.

    See https://www.biosemi.com/faq/file_format.htm for more detailed information about the BDF specification.
    See https://www.teuniz.net/edfbrowser/ for a free EDF/BDF viewer.

    Parameters
    ----------
    data_dir - directory which contains the data to be converted. This location must contain:
        1. a parameter file.
        2. a raw_data.csv file.
        3. a trigger.txt file.
    bdf_path - optional; path to write converted data; defaults to writing a file named raw_data.edf in the data_dir.
        Must end in .edf.
    overwrite - If True, the destination file (if it exists) will be overwritten. If False (default), an error will
        be raised if the file exists.
    write_targetness - If True, and targetness information is available, write that instead of the stimuli markers.
        False by default.
    use_event_durations - optional; if True assigns a duration to each event.
    remove_pre_fixation - optional; if True removes the non-inquiry based markers from the trigger data.
    pre_filter - optional; if True, apply the default filter to the data using to loaded parameters file.

    Returns
    -------
        Path to new bdf file
    """
    if not bdf_path:
        bdf_path = str(Path(data_dir, f'{RAW_DATA_FILENAME}.bdf').resolve())
    if not bdf_path.endswith('.bdf'):
        raise ValueError(f'bdf_path=[{bdf_path}] must end in .bdf')

    data, channels, fs, events, annotation_channels, pre_filter = pyedf_convert(
        data_dir,
        write_targetness=write_targetness,
        use_event_durations=use_event_durations,
        remove_pre_fixation=remove_pre_fixation,
        pre_filter=pre_filter)

    return write_pyedf(
        bdf_path,
        data,
        channels,
        fs,
        events,
        FILETYPE_BDFPLUS,
        overwrite,
        annotation_channels,
        pre_filter)


def pyedf_convert(data_dir: str,
                  write_targetness: bool = False,
                  use_event_durations: bool = False,
                  remove_pre_fixation: bool = False,
                  pre_filter: bool = False) -> Tuple[RawData, List[str], int, List[Tuple[str, int, int]], int]:
    """ Converts BciPy data to formats that can be used by pyEDFlib.

    Parameters
    ----------

    data_dir - directory which contains the data to be converted. This location must contain:
        1. a parameter file.
        2. a raw_data.csv file.
        3. a trigger.txt file.
    write_targetness - If True, and targetness information is available, write that instead of the stimuli markers.
        False by default.
    use_event_durations - optional; if True assigns a duration to each event.
    remove_pre_fixation - optional; if True removes the non-inquiry based markers from the trigger data.
    pre_filter - optional; if True, apply the default filter to the data using to loaded parameters file.

    Returns
    -------
        data - raw data
        channels - list of channel names
        fs - sampling rate
        events - list of events
        annotation_channels - number of annotation channels
        pre_filter - False if no filter was applied,
            otherwise a string of the filter parameters used to filter the data
    """

    params = load_json_parameters(Path(data_dir, DEFAULT_PARAMETER_FILENAME),
                                  value_cast=True)
    data = load_raw_data(Path(data_dir, f'{RAW_DATA_FILENAME}.csv'))
    fs = data.sample_rate
    if pre_filter:
        default_transform = get_default_transform(
            sample_rate_hz=data.sample_rate,
            notch_freq_hz=params.get("notch_filter_frequency"),
            bandpass_low=params.get("filter_low"),
            bandpass_high=params.get("filter_high"),
            bandpass_order=params.get("filter_order"),
            downsample_factor=params.get("down_sampling_rate"),
        )
        raw_data, fs = data.by_channel(transform=default_transform)
        pre_filter = (f"HP:{params.get('filter_low')} LP:{params.get('filter_high')}"
                      f"N:{params.get('notch_filter_frequency')} D:{params.get('down_sampling_rate')} "
                      f"O:{params.get('filter_order')}")
    else:
        raw_data, _ = data.by_channel()
    durations = trigger_durations(params) if use_event_durations else {}
    static_offset = params['static_trigger_offset']
    logger.info(f'Static offset: {static_offset}')

    trigger_type, trigger_timing, trigger_label = trigger_decoder(
        str(Path(data_dir, TRIGGER_FILENAME)),
        remove_pre_fixation=remove_pre_fixation,
        offset=static_offset)

    # validate annotation parameters given data length and trigger count, up the annotation limit byt increasing
    # annotation_channels if needed
    annotation_channels = validate_annotations(len(raw_data[0]) / data.sample_rate, len(trigger_type))

    triggers = compile_triggers(
        trigger_label, trigger_type, trigger_timing, write_targetness)

    events = compile_annotations(triggers, durations)

    return raw_data, data.channels, fs, events, annotation_channels, pre_filter


def validate_annotations(record_time: float, trigger_count: int) -> int:
    """Validate Annotations.

    Using the pyedflib library, it is recommended the number of triggers (or annotations) not exceed the recording
        time in seconds. This may result in an unsuccessful export. The best workaround is to increase the number of
        annotation channels. This function will return the number of annotation channels needed to export the triggers
        successfully. The maximum number of annotation channels is 64 and it is not advised to use more than required
        due to fize size restrictions.

    See https://github.com/holgern/pyedflib/issues/34 for more information.
    """
    if trigger_count > record_time:
        logger.warning(
            f'\n*Warning* The number of triggers [{trigger_count}] exceeds recording time [{record_time}]. ')
        annotation_channels = round(trigger_count / record_time) + 1
        logger.warning(f'\nIncreasing annotation channels to {annotation_channels} to compensate.')
        return annotation_channels
    return 1


def compile_triggers(labels: List[str], targetness: List[str], timing: List[float],
                     write_targetness: bool) -> List[Tuple[str, str, float]]:
    """Compile Triggers.

    Compile trigger information in a way that we edf conversion can easily digest. (label, targetness, timing).
        If write targetness is true, use the targetness as a label.
    """
    triggers = []
    i = 0
    for label in labels:
        # if targetness information available and the flag set to true, write another trigger with target information
        if write_targetness:
            triggers.append((targetness[i], targetness[i], timing[i]))
        else:
            triggers.append((label, targetness[i], timing[i]))
        i += 1
    return triggers


def write_pyedf(output_path: str,
                raw_data: np.array,
                ch_names: List[str],
                sample_rate: float,
                events: List[Tuple[float, float, str]],
                file_type: str = FILETYPE_EDFPLUS,
                overwrite: bool = False,
                annotation_channels: int = 1,
                pre_filter: Optional[str] = None) -> Path:
    """
    Converts BciPy raw_data to the EDF+ or BDF+ filetype using pyEDFlib.

    Adapted from: https://github.com/holgern/pyedflib

    Parameters
    ----------
    output_path - optional path to write converted data; defaults to writing
        a file named raw.edf in the raw_data_dir.
    raw_data - raw data with a row for each channel
    ch_names - names of the channels
    sample_rate - sample frequency
    events - List[Tuple(onset_in_seconds: float, duration_in_seconds: float, description: str)]
    overwrite - If True, the destination file (if it exists) will be overwritten.
        If False (default), an error will be raised if the file exists.
    annotation_channels - number of annotation channels to use
    pre_filter - optional; if provided, add string of filters applied to channel data

    Returns
    -------
        Path to new edf or bdf file
    """
    if not overwrite and os.path.exists(output_path):
        raise OSError(f'{output_path} already exists.')

    physical_min, physical_max = [raw_data.min(), raw_data.max()]
    n_channels = len(raw_data)

    try:
        writer = EdfWriter(output_path, n_channels=n_channels, file_type=file_type)

        channel_info = []
        data_list = []

        for i in range(n_channels):
            ch_dict = {
                'label': ch_names[i],
                'dimension': 'uV',
                'sample_frequency': sample_rate,
                'physical_min': physical_min,
                'physical_max': physical_max,
            }

            if pre_filter:
                ch_dict['prefilter'] = pre_filter

            channel_info.append(ch_dict)
            data_list.append(raw_data[i])

        if events:
            writer.set_number_of_annotation_signals(annotation_channels)
            for onset, duration, label in events:
                writer.writeAnnotation(onset, duration, label)

        writer.setSignalHeaders(channel_info)
        writer.writeSamples(data_list)

    except Exception as error:
        logger.info(error)
        raise error
    finally:
        writer.close()
    return output_path


def compile_annotations(triggers: List[Tuple[str, str, float]],
                        durations: Dict[str, float] = {}
                        ) -> List[Tuple[float, float, str]]:
    """Convert bcipy triggers to the format expected by pyedflib for writing annotations.

    Parameters
    ----------
        triggers - trigger data in the format (symbol, targetness, stamp),
          where stamp has been converted to acquisition clock units.
        durations - optional map defining the duration (seconds) of each
            trigger type. The default is to assign 0.0 seconds.
    Returns
    -------
        List[Tuple(onset_in_seconds, duration_in_seconds, description)]
    """
    return [(timestamp, durations.get(targetness, 0.0), label)
            for (label, targetness, timestamp) in triggers]


def compress(tar_file_name: str, members: List[str]) -> None:
    """
    File compression and archiving.

    Adds files to a tar archive and compresses them using the gzip compression
        format.

    Parameters
    ----------
    tar_file_name (str): name of resulting compressed tar archive. Input string
        can either include or not include file extension (.tar.gz), as this will
        be checked.
    members (List[str]): list of files and folders to be compressed into archive.
        Each file or folder requires the relative file path and full file extension.
        Individual file paths and names that are very long may throw an error
        upon extraction, proceed with caution.

    Returns
    -------
        None
    """

    for member in members:
        # Checks if file exists
        if not os.path.exists(member):
            raise FileNotFoundError(f"This file or folder, '{member}', "
                                    "does not exist!\nPlease rerun program")
        # Checks for file names that may be too long. OS level restriction.
        if len(member) > FILE_LENGTH_LIMIT:
            logger.warning(
                f'File length exceeds compression limit=[{FILE_LENGTH_LIMIT}]. '
                'This may cause issues later with extraction. Please proceed at your own discretion.')

    full_tar_name = tar_name_checker(tar_file_name)
    # Warns user that reopening tar archive that already exists will overwrite it
    if os.path.exists(full_tar_name):
        raise Exception(f"This tar archive=[{full_tar_name}] already exists, continuing will "
                        "overwrite anything in the existing archive.")

    # Opens file for gzip (gz) compressed writing (w)
    # Uses default compression level 9 (highest compression, slowest speed)
    with tarfile.open(full_tar_name, mode="w:gz") as tar:
        # Sets progress bar through tqdm
        progress = tqdm(members)
        for member in progress:
            # Adds file/folder to the tar file and compresses it
            tar.add(member)
            # Sets progress description of progress bar
            progress.set_description(f"Compressing {member}")


def decompress(tar_file: str, path: str) -> None:
    """
    Archive decompression and extraction.

    Takes .tar.gz archive, decompresses, and extracts its contents. Should only be
    used with files created by compress() method.

    Parameters
    ----------
    tar_file (str): name of .tar.gz archive to be decompressed. Input string
        can either include or not include file extension (.tar.gz), as this will
        be checked.
    path (str): file path and name of folder for the extracted contents of the
        archive. If only a name is entered, by default it will place the folder
        in the current working directory. (ex. "extracted" creates a new folder
        in the current directory and extracts the archive contents to it)

    Returns
    -------
        None
    """

    full_tar_name = tar_name_checker(tar_file)

    # Checks if file exists
    if not os.path.exists(full_tar_name):
        raise FileNotFoundError(f"This file or folder, '{tar_file}', "
                                "does not exist!\nPlease rerun program")

    # Opens file for gzip (gz) compressed reading (r)
    with tarfile.open(full_tar_name, mode="r:gz") as tar:
        members = tar.getmembers()

        # Sets progress bar through tqdm
        progress = tqdm(members)
        for member in progress:
            tar.extract(member, path=path)
            # Sets progress description of progress bar
            progress.set_description(f"Extracting {member.name}")


def archive_list(tar_file: str) -> List[str]:
    """
    Returns contents of tar archive.

    Takes .tar.gz archive and returns a full list of its contents, including
        folders, subfolders, and files, with their full relative paths and names.

    Parameters
    ----------
    tar_file (str): name of desired .tar.gz archive. Input string can either
        include or not include file extension (.tar.gz), as this will be checked.

    Returns
    -------
        List[str] of all items in archive
    """

    full_tar_name = tar_name_checker(tar_file)

    # Checks if file exists
    if not os.path.exists(full_tar_name):
        raise FileNotFoundError(f"This file or folder, '{tar_file}', "
                                "does not exist!\nPlease rerun program")

    # Adds names of archive contents to list
    with tarfile.open(full_tar_name, mode="r") as tar:
        tar_list = tar.getnames()
    return tar_list


def tar_name_checker(tar_file_name: str) -> str:
    """
    Checks and modifies file name for tar archive.

    Helper method that takes a tar file name and checks it for the appropriate
        file extension (".tar.gz" for now), returning it if it already does, and
        adding the extension and then returning it if it does not.

    Parameters
    ----------
    tar_file_name (str): name for archive being checked.

    Returns
    -------
        String of properly formatted tar archive name
    """

    if tar_file_name.endswith('.tar.gz'):
        return tar_file_name
    return f'{tar_file_name}.tar.gz'


def convert_to_mne(
        raw_data: RawData,
        channel_map: List[int] = None,
        channel_types: Optional[List[str]] = None,
        transform: Optional[Composition] = None,
        montage: str = 'standard_1020',
        volts: bool = True) -> RawArray:
    """Convert to MNE.

    Returns BciPy RawData as an MNE RawArray. This assumes all channel names
    are reflective of provided montage locations.

    Parameters
    ----------
        raw_data - BciPy RawData object
        channel_map - optional list of channels to include in the MNE RawArray [1, 0, 1, 0, 1, 0, 1, 0].
            1 indicates the channel should be included, 0 indicates it should be excluded.
            Must be the same length as the number of channels in the raw_data.
        channel_types - list of channel types to include in the MNE RawArray.
            If None, all channels will be assumed to be eeg.
            See: https://mne.tools/stable/overview/implementation.html#supported-channel-types
        transform - optional transform to apply to the data
        montage - name of the channel location montage to use.
            See https://mne.tools/dev/generated/mne.channels.make_standard_montage.html
        volts - if True, assume data is already in volts. If false, assume data is in microvolts and convert to volts.
            MNE expects data to be in volts.
            See: https://mne.tools/dev/overview/implementation.html#internal-representation-units
    """
    # if no channel map provided, assume all channels are included
    if not channel_map:
        channel_map = [1] * len(raw_data.channels)

    data, channels, fs = raw_data.by_channel_map(channel_map, transform)

    # if no channel types provided, assume all channels are eeg
    if not channel_types:
        channel_types = ['eeg'] * len(channels)

    # check that number of channel types matches number of channels in the case custom channel types are provided
    assert len(channel_types) == len(channels), \
        f'Number of channel types ({len(channel_types)}) must match number of channels ({len(channels)})'

    info = mne.create_info(channels, fs, channel_types)
    mne_data = RawArray(data, info)
    ten_twenty_montage = mne.channels.make_standard_montage(montage)
    mne_data.set_montage(ten_twenty_montage)

    # convert to volts if necessary (the default for many systems is microvolts)
    if not volts:
        mne_data = mne_data.apply_function(lambda x: x * 1e-6)

    return mne_data


def tobii_to_norm(tobii_units: Tuple[float, float]) -> Tuple[float, float]:
    """Tobii to PsychoPy's 'norm' units.

    https://developer.tobiipro.com/commonconcepts/coordinatesystems.html
    https://www.psychopy.org/general/units.html

    Tobii uses an Active Display Coordinate System.
        The point (0, 0) denotes the upper left corner and (1, 1) the lower right corner of it.

    PsychoPy uses several coordinate systems, the normalized window unit is assumed here.
        The point (0, 0) denotes the center of the screen and (-1, -1) the upper left corner
        and (1, 1) the lower right corner.

    """
    # check that Tobii units are within the expected range
    assert 0 <= tobii_units[0] <= 1, "Tobii x coordinate must be between 0 and 1"
    assert 0 <= tobii_units[1] <= 1, "Tobii y coordinate must be between 0 and 1"

    # convert Tobii units to Psychopy units
    norm_x = (tobii_units[0] - 0.5) * 2
    norm_y = (tobii_units[1] - 0.5) * 2 * -1
    return (norm_x, norm_y)


def norm_to_tobii(norm_units: Tuple[float, float]) -> Tuple[float, float]:
    """PsychoPy's 'norm' units to Tobii.

    https://developer.tobiipro.com/commonconcepts/coordinatesystems.html
    https://www.psychopy.org/general/units.html

    Tobii uses an Active Display Coordinate System.
        The point (0, 0) denotes the upper left corner and (1, 1) the lower right corner of it.

    PsychoPy uses several coordinate systems, the normalized window unit is assumed here.
        The point (0, 0) denotes the center of the screen and (-1, -1) the upper left corner
        and (1, 1) the lower right corner.
    """
    # check that the coordinates are within the bounds of the screen
    assert norm_units[0] >= -1 and norm_units[0] <= 1, "X coordinate must be between -1 and 1"
    assert norm_units[1] >= -1 and norm_units[1] <= 1, "Y coordinate must be between -1 and 1"

    # convert PsychoPy norm units to Tobii units
    tobii_x = (norm_units[0] / 2) + 0.5
    tobii_y = ((norm_units[1] * -1) / 2) + 0.5
    return (tobii_x, tobii_y)



import logging
from typing import List, Tuple, Optional
from bcipy.display import (
    BCIPY_LOGO_PATH,
    Display,
    InformationProperties,
    VEPStimuliProperties,
    TaskDisplayProperties,
)
import numpy as np
from psychopy import visual, core
from bcipy.helpers.clock import Clock
from bcipy.helpers.symbols import SPACE_CHAR
from bcipy.helpers.triggers import TriggerCallback, _calibration_trigger
from bcipy.helpers.stimuli import resize_image, get_fixation


def create_vep_codes(length=32, count=4) -> List[List[int]]:
    """Create a list of random VEP codes.

    length - how many bits in each code. This should be greater than or equal to the refresh rate
        if using these to flicker. For example, if the refresh rate is 60Hz, then the length should
        be at least 60.
    count - how many codes to generate, each will be unique.
    """
    np.random.seed(1)
    return [np.random.randint(2, size=length) for _ in range(count)]


class VEPBox:
    """A class to represent a VEP box.

    Attributes:
        flicker: A list of two visual.GratingStim objects, one for on and one for off.
        code: A list of integers representing the VEP code for this box.
    """

    def __init__(self, flicker: List[Tuple[visual.GratingStim, visual.GratingStim]], code: List[int]) -> None:
        self.flicker = flicker
        self.code = code

    def frame_on(self) -> None:
        """Frame On. Draw the frame for the first grating stim on the screen."""
        self.flicker[0].draw()

    def frame_off(self) -> None:
        """Frame Off. Draw the frame for the second grating stim on the screen."""
        self.flicker[1].draw()


class VEPDisplay(Display):

    def __init__(
            self,
            window: visual.Window,
            experiment_clock: Clock,
            stimuli: VEPStimuliProperties,
            task_display: TaskDisplayProperties,
            info: InformationProperties,
            trigger_type: str = 'text',
            space_char: str = SPACE_CHAR,
            full_screen: bool = False,
            codes: List[int] = None):
        self.window = window
        self.window_size = self.window.size  # [w, h]
        self.refresh_rate = round(window.getActualFrameRate())

        self.logger = logging.getLogger(__name__)

        # Stimuli parameters, these are set on display in order to allow
        #  easy updating after defintion
        self.stimuli_inquiry = stimuli.stim_inquiry
        self.stimuli_colors = stimuli.stim_colors
        self.stimuli_timing = stimuli.stim_timing
        self.stimuli_font = stimuli.stim_font
        self.stimuli_height = stimuli.stim_height
        self.stimuli_pos = stimuli.stim_pos
        self.logger.info(self.stimuli_pos)
        self.is_txt_stim = stimuli.is_txt_stim
        self.stim_length = stimuli.stim_length
        self.sti = []
        self.fixation = visual.TextStim(
            self.window,
            text=get_fixation(True),
            color='red',
            height=self.stimuli_height,
            units='height',
            pos=[0, 0])

        self.full_screen = full_screen

        self.static_clock = core.Clock()

        # Trigger handling
        self.first_stim_time = None
        self.trigger_type = trigger_type
        self.trigger_callback = TriggerCallback()
        self.experiment_clock = experiment_clock

        # Callback used on presentation of first stimulus.
        self.first_run = True
        self.first_stim_callback = lambda _sti: None
        self.size_list_sti = []

        # Task parameters
        self.space_char = space_char
        self.task_display = task_display
        self.task = task_display.build_task(self.window)

        # Information parameters
        self.info = info
        self.info_text = info.build_info_text(window)

        self.colors = [('white', 'black'), ('red', 'green'), ('blue', 'yellow'), ('orange', 'green')]
        if not codes:
            self.codes = create_vep_codes(length=self.refresh_rate, count=4)
        else:
            self.codes = codes

        # build the VEP stimuli
        self.vep = []
        self.text_boxes = []
        self.vep_type = 4
        assert self.vep_type == 4, 'Only 4 stimuli are supported in VEP display at this time!'
        self.vep_box_size = .1

        self.animation_duration = 2

    def do_fixation(self) -> None:
        # draw fixation cross
        self.fixation.draw()
        self.draw_static()
        self.window.flip()
        core.wait(self.stimuli_timing[1])

    def do_inquiry(self) -> List[float]:
        """Do the inquiry."""

        timing = []

        # if this is the first run, calibrate using the trigger pulse and setup the VEP and text boxes
        if self.first_run:
            self._trigger_pulse()
            self._build_vep_corner_stimuli(self.stimuli_pos, self.codes, self.colors, self.vep_box_size)
            self._build_text_boxes()

        # fixation --> animation / prompting --> VEP stimulate
        self.do_fixation()
        timing += self.animate_inquiry()
        timing += self.stimulate()

        # clear everyting expect static stimuli
        self.draw_static()
        self.window.flip()

        return timing

    def animate_inquiry(self) -> List[float]:
        """Display the inquiry.

        Inquiry is a list of lists of strings.
        Each list contains what stimuli to display for each box defined in self.vep.
        """
        timing = []
        self.trigger_callback.reset()
        self.window.callOnFlip(
            self.trigger_callback.callback,
            self.experiment_clock,
            'VEP_INQ_ANIMATION')

        self._reset_text_boxes()
        self._build_inquiry_stimuli()

        self.static_clock.reset()
        while self.static_clock.getTime() < self.animation_duration:
            self.draw_boxes()
            self.draw_animation()
            self.draw_static()
            self.window.flip()
        timing += self.trigger_callback.timing

        self._set_inquiry()
        self.trigger_callback.reset()
        self.window.callOnFlip(
            self.trigger_callback.callback,
            self.experiment_clock,
            'VEP_INQUIRY')
        # display the inquiry
        self.static_clock.reset()
        while self.static_clock.getTime() < self.stimuli_timing[0]:
            self.draw_boxes()
            self.draw_static()
            self.window.flip()

        timing += self.trigger_callback.timing

        return timing

    def draw_animation(self) -> None:
        """Draw the stimuli animation."""
        for sti in self.sti:
            sti.draw()

    def draw_boxes(self) -> None:
        """Draw the text boxes under VEP stimuli."""
        for box in self.text_boxes:
            box.draw()

    def stimulate(self) -> List[float]:
        """Stimulate.

        This is the main display function of the VEP paradigm. It is responsible for drawing the flickering stimuli.

        It assumes that the VEP stimuli are already constructed using the _build_vep_corner_stimuli function.
        These boxes are drawn in the order they are in the list as defined in self.vep."""
        self.trigger_callback.reset()
        self.window.callOnFlip(
            self.trigger_callback.callback,
            self.experiment_clock,
            'VEP_STIMULATE')
        self.static_clock.reset()
        while self.static_clock.getTime() < self.stimuli_timing[-1]:
            for frame in range(self.refresh_rate):
                for box in self.vep:
                    if box.code[frame] == 1:
                        box.frame_on()
                    else:
                        box.frame_off()

                # flicker!
                self.draw_static()
                self.window.flip()

        return self.trigger_callback.timing

    def draw_static(self) -> None:
        """Draw static elements for the display."""
        self.task.draw()
        for info in self.info_text:
            info.draw()

    def update_task(self, text: str, color: str, pos: Optional[Tuple] = None) -> None:
        """Update the task display which is displayed using the draw_static function."""
        self.task.text = text
        self.task.color = color
        if pos:
            self.task.pos = pos

    def _build_inquiry_stimuli(self) -> None:
        """Build the inquiry stimuli."""
        all_letters = []
        all_colors = []
        self.sti = []
        for letters, colors in zip(self.stimuli_inquiry, self.stimuli_colors):
            for letter in letters:
                all_letters.append(letter)
                all_colors.append(colors[0])
        pos = (0, 0)
        for letter, color in zip(all_letters, all_colors):
            pos = (pos[0] + self.stimuli_height, pos[1])
            self.sti.append(
                visual.TextStim(self.window, text=letter, color=color, pos=pos, height=self.stimuli_height))

        return self.sti

    def _build_vep_corner_stimuli(
            self,
            pos: Tuple[float, float],
            codes: List[int],
            colors: List[Tuple[str]],
            box_size: float) -> List[visual.GratingStim]:
        """Build the corner stimuli for the VEP.

        Args:
            window: psychopy window object
        """
        corner_stim = []
        for pos, code, color in zip(pos, codes, colors):
            corner_stim += self._build_2x2_vep_grid(pos, box_size, color, code)

        return corner_stim

    def _build_1x1_vep_grid(
            self,
            position: Tuple[float],
            size: float,
            color: str,
            code: List[int]) -> List[visual.GratingStim]:
        """Build a 1x1 VEP grid.

        Args:
            position: position of the flicking box
            size: size of the flicking box
            color: color of the flicking box
            code: code to be used with the flicking box

        Returns:
            list of visual.GratingStim objects

        """
        assert len(code) >= self.refresh_rate, 'Code must be longer than refresh rate'
        pattern1 = visual.GratingStim(win=self.window, name=f'1x1-1-{position}', units='height',
                                      tex=None, mask=None,
                                      ori=0, pos=position, size=size, sf=1, phase=0.0,
                                      color=color[0], colorSpace='rgb', opacity=1,
                                      texRes=256, interpolate=True, depth=-1.0)
        pattern2 = visual.GratingStim(win=self.window, name=f'1x1-2-{position}', units='height',
                                      tex=None, mask=None,
                                      ori=0, pos=position, size=size, sf=1, phase=0,
                                      color=color[1], colorSpace='rgb', opacity=1,
                                      texRes=256, interpolate=True, depth=-2.0)
        box = [VEPBox(flicker=[pattern1, pattern2], code=code)]
        self.vep += box
        return box

    def _build_2x2_vep_grid(self,
                            position: Tuple[float],
                            size: float,
                            color: str,
                            code: List[int]) -> List[visual.GratingStim]:
        """Build a 2x2 VEP grid.

        Args:
            position: position of the flicking box
            size: size of the flicking box
            color: color of the flicking box
            code: code to be used with the flicking box

        Returns:
            list of visual.GratingStim objects
        """

        assert len(code) >= self.refresh_rate, 'Code must be longer than refresh rate'
        starting_position = position
        positions = [
            starting_position,  # bottom left
            [starting_position[0] + size, starting_position[1]],  # bottom right
            [starting_position[0] + size, starting_position[1] + size],  # top right
            [starting_position[0], starting_position[1] + size],  # top left
        ]
        box = []
        for pos in positions:
            pattern1 = visual.GratingStim(win=self.window, name=f'2x2-1-{pos}', units='height',
                                          tex=None, mask=None,
                                          ori=0, pos=pos, size=size, sf=1, phase=0.0,
                                          color=color[0], colorSpace='rgb', opacity=1,
                                          texRes=256, interpolate=True, depth=-1.0)
            pattern2 = visual.GratingStim(win=self.window, name=f'2x2-2-{pos}', units='height',
                                          tex=None, mask=None,
                                          ori=0, pos=pos, size=size, sf=1, phase=0,
                                          color=color[1], colorSpace='rgb', opacity=1,
                                          texRes=256, interpolate=True, depth=-2.0)
            color = (color[1], color[0])
            box += [VEPBox(flicker=[pattern1, pattern2], code=code)]

        self.vep += box
        return box

    def _trigger_pulse(self) -> None:
        """Trigger Pulse.

        This method uses a calibration trigger to determine any functional
            offsets needed for operation with this display. By setting the first_stim_time and searching for the
            same stimuli output to the marker stream, the offsets between these proceses can be reconciled at the
            beginning of an experiment. If drift is detected in your experiment, more frequent pulses and offset
            correction may be required.
        """
        calibration_time = _calibration_trigger(
            self.experiment_clock,
            trigger_type=self.trigger_type,
            display=self.window)

        # set the first stim time if not present and first_run to False
        if not self.first_stim_time:
            self.first_stim_time = calibration_time[-1]
            self.first_run = False

    def schedule_to(self, stimuli: List[List[str]], timing: List[List[float]], colors: List[List[str]]) -> None:
        """Schedule stimuli elements (works as a buffer).
        """
        self.stimuli_inquiry = stimuli
        self.stimuli_timing = timing
        self.stimuli_colors = colors

    def _build_text_boxes(self) -> List[visual.TextBox2]:
        """Build the text boxes for the experiment.

        This assumes that vep_type is 4 using a 2x2 grid. Additional configurations can be added in the future.
        """
        # generate the inquiry stimuli
        assert len(
            self.stimuli_colors) == self.vep_type, (
                f"stmuli colors {len(self.stimuli_colors)} must be the same length as vep type {self.vep_type}")
        assert len(
            self.stimuli_pos) == self.vep_type, (
                f"stmuli position {len(self.stimuli_pos)} must be the same length as vep type {self.vep_type}")

        self.text_boxes = []
        inc_q = int(self.vep_type / 2)
        size = self.vep_box_size * inc_q
        inc = 0  # get the last box in the grating stim per vep
        for color in self.stimuli_colors:
            text = ' '
            first = self.vep[inc].flicker[0].pos
            best_neighbor = self.vep[inc + inc_q].flicker[0].pos

            # the textbox requires the pos to be centered on the middle of the box, so
            # we need to calculate the center of the vep stimulus boxes
            pos = (first[0] + best_neighbor[0]) / 2, (first[1] + best_neighbor[1]) / 2
            self.text_boxes += [visual.TextBox2(
                win=self.window,
                text=text,
                font=self.stimuli_font,
                units='height',
                pos=pos,
                color=color[0],
                colorSpace='rgb',
                size=[size, size],
                alignment='center',
                anchor='center',
                borderWidth=2,
                borderColor='white',
                letterHeight=self.stimuli_height / 2)]
            inc += self.vep_type
        return self.text_boxes

    def _reset_text_boxes(self) -> None:
        """Reset text boxes.

        This method resets the text boxes to the blank state.
        """
        for text_box in self.text_boxes:
            text_box.setText(' ')

    def _set_inquiry(self) -> List[visual.TextBox2]:
        """Generate Inquiry.

        This method sets the correct inquiry text for each text boxes.
        """
        # generate the inquiry stimuli
        assert len(
            self.stimuli_inquiry) == self.vep_type, (
                f"stmuli inquiry {len(self.stimuli_inquiry)} must be the same length as vep type {self.vep_type}")
        assert len(
            self.stimuli_colors) == self.vep_type, (
                f"stmuli colors {len(self.stimuli_colors)} must be the same length as vep type {self.vep_type}")
        for box_content, color, box in zip(self.stimuli_inquiry, self.stimuli_colors, self.text_boxes):
            text = ' '.join(box_content)
            box.text = text
            box.color = color[0]
        return self.text_boxes

    def wait_screen(self, message: str, color: str) -> None:
        """Wait Screen.

        Args:
            message(string): message to be displayed while waiting
            color(string): color of the message to be displayed
        """

        # Construct the wait message
        wait_message = visual.TextStim(win=self.window,
                                       font=self.stimuli_font,
                                       text=message,
                                       height=.1,
                                       color=color,
                                       pos=(0, -.5),
                                       wrapWidth=2,
                                       colorSpace='rgb',
                                       opacity=1,
                                       depth=-6.0)

        # try adding the BciPy logo to the wait screen
        try:
            wait_logo = visual.ImageStim(
                self.window,
                image=BCIPY_LOGO_PATH,
                pos=(0, .25),
                mask=None,
                ori=0.0)
            wait_logo.size = resize_image(
                BCIPY_LOGO_PATH,
                self.window.size,
                1)
            wait_logo.draw()

        except Exception as e:
            self.logger.exception(f'Cannot load logo image from path=[{BCIPY_LOGO_PATH}]')
            raise e

        # Draw and flip the screen.
        wait_message.draw()
        self.draw_static()
        self.window.flip()



import numpy as np


def truncate_float(num, precision):
    return float(str(num)[:precision])


def gen_random_data(low, high, channel_count, precision=8):
    return [truncate_float(np.random.uniform(low, high), precision)
            for _ in range(channel_count)]



import pickle
from pathlib import Path
from typing import List

import numpy as np
from bcipy.signal.model import ModelEvaluationReport, SignalModel
from bcipy.signal.model.classifier import RegularizedDiscriminantAnalysis
from bcipy.signal.model.cross_validation import cost_cross_validation_auc, cross_validation
from bcipy.signal.model.density_estimation import KernelDensityEstimate
from bcipy.signal.model.dimensionality_reduction import MockPCA
from bcipy.signal.model.pipeline import Pipeline
from bcipy.helpers.exceptions import SignalException
from bcipy.helpers.stimuli import InquiryReshaper


class RdaKdeModel(SignalModel):
    reshaper = InquiryReshaper()

    def __init__(self, k_folds: int, prior_type: str = "uniform"):
        self.k_folds = k_folds
        self.prior_type = prior_type
        self.optimization_elements = 1  # number of elements to optimized (RDA)
        self.min = 1e-2  # min and max values for the likelihood ratio output
        self.max = 1e2
        self.model = None
        self.auc = None
        self._ready_to_predict = False

    def fit(self, train_data: np.array, train_labels: np.array) -> SignalModel:
        """
        Train on provided data using K-fold cross validation and return self.

        Parameters:
            train_data: shape (Channels, Trials, Trial_length) preprocessed data
            train_labels: shape (Trials,) binary labels

        Returns:
            trained likelihood model
        """
        model = Pipeline([MockPCA(), RegularizedDiscriminantAnalysis()])

        # Find the optimal gamma + lambda values
        arg_cv = cross_validation(train_data, train_labels, model=model, k_folds=self.k_folds)

        # Get the AUC using those optimized gamma + lambda
        rda_index = 1  # the index in the pipeline
        model.pipeline[rda_index].lam = arg_cv[0]
        model.pipeline[rda_index].gam = arg_cv[1]
        tmp, sc_cv, y_cv = cost_cross_validation_auc(
            model, rda_index, train_data, train_labels, arg_cv, k_folds=self.k_folds, split="uniform"
        )
        self.auc = -tmp

        # After finding cross validation scores do one more round to learn the
        # final RDA model
        model.fit(train_data, train_labels)

        # Insert the density estimates to the model and train using the cross validated
        # scores to avoid over fitting. Observe that these scores are not obtained using
        # the final model
        model.add(KernelDensityEstimate(scores=sc_cv))
        model.pipeline[-1].fit(sc_cv, y_cv)

        self.model = model

        if self.prior_type == "uniform":
            self.log_prior_class_1 = self.log_prior_class_0 = np.log(0.5)
        elif self.prior_type == "empirical":
            prior_class_1 = np.sum(train_labels == 1) / len(train_labels)
            self.log_prior_class_1 = np.log(prior_class_1)
            self.log_prior_class_0 = np.log(1 - prior_class_1)
        else:
            raise ValueError("prior_type must be 'empirical' or 'uniform'")

        self._ready_to_predict = True
        return self

    def evaluate(self, test_data: np.array, test_labels: np.array) -> ModelEvaluationReport:
        """Computes AUROC of the intermediate RDA step of the pipeline using k-fold cross-validation

        Args:
            test_data (np.array): shape (Channels, Trials, Trial_length) preprocessed data.
            test_labels (np.array): shape (Trials,) binary labels.

        Raises:
            SignalException: error if called before model is fit.

        Returns:
            ModelEvaluationReport: stores AUC
        """
        if not self._ready_to_predict:
            raise SignalException("must use model.fit() before model.evaluate()")

        tmp_model = Pipeline([self.model.pipeline[0], self.model.pipeline[1]])

        lam_gam = (self.model.pipeline[1].lam, self.model.pipeline[1].gam)
        tmp, _, _ = cost_cross_validation_auc(
            tmp_model, self.optimization_elements, test_data, test_labels,
            lam_gam, k_folds=self.k_folds, split="uniform"
        )
        auc = -tmp
        return ModelEvaluationReport(auc)

    def predict(self, data: np.array, inquiry: List[str], symbol_set: List[str]) -> np.array:
        """
        For each trial in `data`, compute a likelihood ratio to update that symbol's probability.
        Rather than just computing an update p(e|l=+) for the seen symbol and p(e|l=-) for all unseen symbols,
        we compute a likelihood ratio p(e | l=+) / p(e | l=-) to update the seen symbol, and all other symbols
        can receive a multiplicative update of 1.

        Args:
            data (np.array): EEG data with shape (n_channel, n_trial, n_sample).
            inquiry (List[str]): List describing the symbol shown in each trial.
            symbol_set (List[str]): The set of all possible symbols.

        Raises:
            SignalException: error if called before model is fit.

        Returns:
            np.array: multiplicative update term (likelihood ratios) for each symbol in the `symbol_set`.
        """

        if not self._ready_to_predict:
            raise SignalException("must use model.fit() before model.predict()")

        # Evaluate likelihood probabilities for p(e|l=1) and p(e|l=0)
        log_likelihoods = self.model.transform(data)
        subset_likelihood_ratios = np.exp(log_likelihoods[:, 1] - log_likelihoods[:, 0])
        # Restrict multiplicative updates to a reasonable range
        subset_likelihood_ratios = np.clip(subset_likelihood_ratios, self.min, self.max)

        # Apply likelihood ratios to entire symbol set.
        likelihood_ratios = np.ones(len(symbol_set))
        for idx in range(len(subset_likelihood_ratios)):
            likelihood_ratios[symbol_set.index(inquiry[idx])] *= subset_likelihood_ratios[idx]
        return likelihood_ratios

    def predict_proba(self, data: np.array) -> np.array:
        """Converts log likelihoods from model into class probabilities.

        Returns:
            posterior (np.ndarray): shape (num_items, 2) - for each item, the model's predicted
                probability for the two labels.
        """
        if not self._ready_to_predict:
            raise SignalException("must use model.fit() before model.predict_proba()")

        # Model originally produces p(eeg | label). We want p(label | eeg):
        #
        # p(l=1 | e) = p(e | l=1) p(l=1) / p(e)
        # log(p(l=1 | e)) = log(p(e | l=1)) + log(p(l=1)) - log(p(e))
        log_scores_class_0 = self.model.transform(data)[:, 0]
        log_scores_class_1 = self.model.transform(data)[:, 1]
        log_post_0 = log_scores_class_0 + self.log_prior_class_0
        log_post_1 = log_scores_class_1 + self.log_prior_class_1
        denom = np.logaddexp(log_post_0, log_post_1)
        log_post_0 -= denom
        log_post_1 -= denom
        posterior = np.exp(np.stack([log_post_0, log_post_1], axis=-1))
        return posterior

    def save(self, path: Path):
        """Save model weights (e.g. after training) to `path`"""
        with open(path, "wb") as f:
            pickle.dump(self.model, f)

    def load(self, path: Path):
        """Load pretrained model weights from `path`"""
        with open(path, "rb") as f:
            self.model = pickle.load(f)
        self._ready_to_predict = True



import argparse
import logging
import multiprocessing

from typing import List

from psychopy import visual

from bcipy.acquisition import LslDataServer, ClientManager
from bcipy.display import init_display_window
from bcipy.config import DEFAULT_PARAMETERS_PATH, DEFAULT_EXPERIMENT_ID, STATIC_AUDIO_PATH
from bcipy.helpers.acquisition import init_eeg_acquisition
from bcipy.helpers.language_model import init_language_model
from bcipy.helpers.load import (load_experiments, load_json_parameters,
                                load_signal_model)
from bcipy.helpers.save import init_save_data_structure
from bcipy.helpers.session import collect_experiment_field_data
from bcipy.helpers.stimuli import play_sound
from bcipy.helpers.system_utils import configure_logger, get_system_info
from bcipy.helpers.task import print_message
from bcipy.helpers.validate import validate_experiment, validate_bcipy_session
from bcipy.helpers.visualization import visualize_session_data
from bcipy.signal.model import PcaRdaKdeModel
from bcipy.task import TaskType
from bcipy.task.start_task import start_task

log = logging.getLogger(__name__)


def bci_main(
        parameter_location: str,
        user: str,
        task: TaskType,
        experiment: str = DEFAULT_EXPERIMENT_ID,
        alert: bool = False,
        visualize: bool = True,
        fake: bool = False) -> bool:
    """BCI Main.

    The BCI main function will initialize a save folder, construct needed information
    and execute the task. This is the main connection between any UI and
    running the app.

    It may also be invoked via tha command line.
        Ex. `bcipy` this will default parameters, mode, user, and type.

        You can pass it those attributes with flags, if desired.
            Ex. `bcipy --user "bci_user" --task "RSVP Calibration" --experiment "default"

    Input:
        parameter_location (str): location of parameters file to use
        user (str): name of the user
        task (TaskType): registered bcipy TaskType
        experiment_id (str): Name of the experiment. Default name is DEFAULT_EXPERIMENT_ID.
        alert (bool): whether to alert the user when the task is complete
        visualize (bool): whether to visualize data at the end of a task
        fake (bool): whether to use fake acquisition data during the session. If None, the
            fake data will be determined by the parameters file.
    """
    validate_experiment(experiment)
    # Load parameters
    parameters = load_json_parameters(parameter_location, value_cast=True)

    # cli overrides parameters file for fake data if provided
    fake = fake if fake is True else parameters['fake_data']

    if not validate_bcipy_session(parameters, fake):
        return False

    # Update property to reflect the parameter source
    parameters['parameter_location'] = parameter_location
    if parameter_location != DEFAULT_PARAMETERS_PATH:
        parameters.save()
        default_params = load_json_parameters(DEFAULT_PARAMETERS_PATH, value_cast=True)
        if parameters.add_missing_items(default_params):
            msg = 'Parameters file out of date.'
            log.exception(msg)
            raise Exception(msg)

    # update our parameters file with system related information
    sys_info = get_system_info()

    # Initialize Save Folder
    save_folder = init_save_data_structure(
        parameters['data_save_loc'],
        user,
        parameter_location,
        task=task.label,
        experiment_id=experiment)

    # configure bcipy session logging
    configure_logger(save_folder,
                     version=sys_info['bcipy_version'])

    log.info(sys_info)

    # Collect experiment field data
    collect_experiment_field_data(experiment, save_folder)

    if execute_task(task, parameters, save_folder, alert, fake):
        if visualize:
            visualize_session_data(save_folder, parameters)
        return True

    return False


def execute_task(
        task: TaskType,
        parameters: dict,
        save_folder: str,
        alert: bool,
        fake: bool) -> bool:
    """Execute Task.

    Executes the desired task by setting up the display window and
        data acquisition, then passing on to the start_task function
        which will initialize experiment.

    Input:
        (str): registered bcipy TaskType
        parameters (dict): parameter dictionary
        save_folder (str): path to save folder
        alert (bool): whether to alert the user when the task is complete
        fake (bool): whether to use fake acquisition data during the session

    Returns:
        (bool): True if the task was successfully executed, False otherwise
    """
    signal_model = None
    language_model = None

    # Init EEG Model, if needed. Calibration Tasks Don't require probabilistic
    # modules to be loaded.
    if task not in TaskType.calibration_tasks():
        # Try loading in our signal_model and starting a langmodel(if enabled)
        if not fake:
            try:
                signal_model, _filename = load_signal_model(
                    model_class=PcaRdaKdeModel,
                    model_kwargs={'k_folds': parameters['k_folds']},
                    filename=parameters['signal_model_path'])
            except Exception as e:
                log.exception(f'Cannot load signal model. Exiting. {e}')
                raise e

        language_model = init_language_model(parameters)

    # Initialize DAQ and export the device configuration
    daq, servers = init_eeg_acquisition(
        parameters, save_folder, server=fake)

    # Initialize Display Window
    # We have to wait until after the prompt to load the signal model before
    # displaying the window, otherwise in fullscreen mode this throws an error
    display = init_display_window(parameters)
    print_message(display, f'Initializing {task}...')

    # Start Task
    try:
        start_task(display,
                   daq.get_client('EEG'),
                   task,
                   parameters,
                   save_folder,
                   language_model=language_model,
                   signal_model=signal_model,
                   fake=fake)

    # If exception, close all display and acquisition objects
    except Exception as e:
        log.exception(str(e))

    if alert:
        play_sound(f"{STATIC_AUDIO_PATH}/{parameters['alert_sound_file']}")

    return _clean_up_session(display, daq, servers)


def _clean_up_session(
        display: visual.Window,
        daq: ClientManager,
        servers: List[LslDataServer] = None) -> bool:
    """Clean up session.

    Closes the display window and data acquisition objects. Returns True if the session was closed successfully.

    Input:
        display (visual.Window): display window
        daq (LslAcquisitionClient): data acquisition client
        server (LslDataServer): data server
    """
    try:
        # Stop Acquisition
        daq.stop_acquisition()
        daq.cleanup()

        for server in servers:
            server.stop()

        # Close the display window
        # NOTE: There is currently a bug in psychopy when attempting to shutdown
        # windows when using a USB-C monitor. Putting the display close last in
        # the inquiry allows acquisition to properly shutdown.
        display.close()
        return True
    except Exception as e:
        log.exception(str(e))
        return False


def bcipy_main() -> None:
    """BciPy Main.

    Command line interface used for running a registered experiment task in BciPy. To see what
        is available use the --help flag.
    """
    # Needed for windows machines
    multiprocessing.freeze_support()

    experiment_options = list(load_experiments().keys())
    task_options = TaskType.list()
    parser = argparse.ArgumentParser()

    # Command line utility for adding arguments/ paths via command line
    parser.add_argument('-p', '--parameters', default=DEFAULT_PARAMETERS_PATH,
                        help='Parameter location. Pass as *.json')
    parser.add_argument('-u', '--user', default='test_user')
    parser.add_argument('-t', '--task', default='RSVP Calibration',
                        help=f'Task type to execute. Registered options: {task_options}')
    parser.add_argument(
        '-e',
        '--experiment',
        default=DEFAULT_EXPERIMENT_ID,
        help=f'Select a valid experiment to run the task for this user. Available options: {experiment_options}')
    parser.add_argument(
        '-a',
        '--alert',
        default=False,
        action='store_true',
        help='Alert the user when the session is complete.')
    parser.add_argument(
        '-nv',
        '--noviz',
        default=True,
        action='store_false',
        help='Suppress visuals of the data after the session is complete.')
    parser.add_argument(
        '-f',
        '--fake',
        default=False,
        action='store_true',
        help='Use fake acquisition data for testing.')
    args = parser.parse_args()

    # Start BCI Main
    bci_main(args.parameters, str(args.user), TaskType.by_value(str(args.task)),
             str(args.experiment), args.alert, args.noviz, args.fake)


if __name__ == '__main__':
    bcipy_main()



import json
import logging
import os
from pathlib import Path
from shutil import copyfile
from time import localtime, strftime
from typing import Any, Dict, List, Tuple

from bcipy.config import (
    ROOT,
    DEFAULT_ENCODING,
    DEFAULT_EXPERIMENT_PATH,
    DEFAULT_PARAMETERS_PATH,
    DEFAULT_FIELD_PATH,
    EXPERIMENT_FILENAME,
    FIELD_FILENAME)
from bcipy.gui.file_dialog import ask_directory, ask_filename
from bcipy.preferences import preferences
from bcipy.helpers.exceptions import (BciPyCoreException,
                                      InvalidExperimentException)
from bcipy.helpers.parameters import Parameters
from bcipy.helpers.raw_data import RawData
from bcipy.signal.model import SignalModel

log = logging.getLogger(__name__)


def copy_parameters(path: str = DEFAULT_PARAMETERS_PATH,
                    destination: str = None) -> str:
    """Creates a copy of the given configuration (parameters.json) to the
    given directory and returns the path.

    Parameters:
    -----------
        path: str - optional path of parameters file to copy; used default if not provided.
        destination: str - optional destination directory; default is the same
          directory as the default parameters.
    Returns:
    --------
        path to the new file.
    """
    default_dir = str(Path(DEFAULT_PARAMETERS_PATH).parent)

    destination = default_dir if destination is None else destination
    filename = strftime('parameters_%Y-%m-%d_%Hh%Mm%Ss.json', localtime())

    path = str(Path(destination, filename))
    copyfile(DEFAULT_PARAMETERS_PATH, path)
    return path


def load_experiments(path: str = f'{DEFAULT_EXPERIMENT_PATH}/{EXPERIMENT_FILENAME}') -> dict:
    """Load Experiments.

    PARAMETERS
    ----------
    :param: path: string path to the experiments file.

    Returns
    -------
        A dictionary of experiments, with the following format:
            { name: { fields : {name: '', required: bool, anonymize: bool}, summary: '' } }

    """
    with open(path, 'r', encoding=DEFAULT_ENCODING) as json_file:
        return json.load(json_file)


def extract_mode(bcipy_data_directory: str) -> str:
    """Extract Mode.

    This method extracts the task mode from a BciPy data save directory. This is important for
        trigger conversions and extracting targeteness.

    *note*: this is not compatible with older versions of BciPy (pre 1.5.0) where
        the tasks and modes were considered together using integers (1, 2, 3).

    PARAMETERS
    ----------
    :param: bcipy_data_directory: string path to the data directory
    """
    directory = bcipy_data_directory.lower()
    if 'calibration' in directory:
        return 'calibration'
    elif 'copy' in directory:
        return 'copy_phrase'
    raise BciPyCoreException(f'No valid mode could be extracted from [{directory}]')


def load_fields(path: str = f'{DEFAULT_FIELD_PATH}/{FIELD_FILENAME}') -> dict:
    """Load Fields.

    PARAMETERS
    ----------
    :param: path: string path to the fields file.

    Returns
    -------
        A dictionary of fields, with the following format:
            {
                "field_name": {
                    "help_text": "",
                    "type": ""
            }

    """
    with open(path, 'r', encoding=DEFAULT_ENCODING) as json_file:
        return json.load(json_file)


def load_experiment_fields(experiment: dict) -> list:
    """Load Experiment Fields.

    {
        'fields': [{}, {}],
        'summary': ''
    }

    Using the experiment dictionary, loop over the field keys and put them in a list.
    """
    if isinstance(experiment, dict):
        try:
            return [name for field in experiment['fields'] for name in field.keys()]
        except KeyError:
            raise InvalidExperimentException(
                'Experiment is not formatted correctly. It should be passed as a dictionary with the fields and'
                f' summary keys. Fields is a list of dictionaries. Summary is a string. \n experiment=[{experiment}]')
    raise TypeError('Unsupported experiment type. It should be passed as a dictionary with the fields and summary keys')


def load_json_parameters(path: str, value_cast: bool = False) -> Parameters:
    """Load JSON Parameters.

    Given a path to a json of parameters, convert to a dictionary and optionally
        cast the type.

    Expects the following format:
    "fake_data": {
        "value": "true",
        "section": "bci_config",
        "readableName": "Fake Data Sessions",
        "helpTip": "If true, fake data server used",
        "recommended_values": "",
        "type": "bool"
        }

    PARAMETERS
    ----------
    :param: path: string path to the parameters file.
    :param: value_case: True/False cast values to specified type.

    Returns
    -------
        a Parameters object that behaves like a dict.
    """
    return Parameters(source=path, cast_values=value_cast)


def load_experimental_data() -> str:
    filename = ask_directory()  # show dialog box and return the path
    log.debug("Loaded Experimental Data From: %s" % filename)
    return filename


def load_signal_model(model_class: SignalModel,
                      model_kwargs: Dict[str, Any], filename: str = None) -> Tuple[SignalModel, str]:
    """Construct the specified model and load pretrained parameters.

    Args:
        model_class (SignalModel, optional): Model class to construct.
        model_kwargs (dict, optional): Keyword arguments for constructing model.
        filename (str, optional): Location of pretrained model parameters.

    Returns:
        SignalModel: Model after loading pretrained parameters.
    """
    # use python's internal gui to call file explorers and get the filename
    if not filename or Path(filename).is_dir():
        directory = filename or preferences.signal_model_directory
        filename = ask_filename('*.pkl', directory)

        # update preferences
        path = Path(filename)
        if path.is_file():
            preferences.signal_model_directory = str(path.parent)

    # load the signal_model with pickle
    signal_model = model_class(**model_kwargs)
    signal_model.load(filename)

    log.info(f'Loaded signal model from {filename}')

    return signal_model, filename


def choose_csv_file(filename: str = None) -> str:
    """GUI prompt to select a csv file from the file system.

    Parameters
    ----------
    - filename : optional filename to use; if provided the GUI is not shown.

    Returns
    -------
    file name of selected file; throws an exception if the file is not a csv.
    """
    if not filename:
        filename = ask_filename('*.csv')

    # get the last part of the path to determine file type
    file_name = filename.split('/')[-1]

    if 'csv' not in file_name:
        raise Exception(
            'File type unrecognized. Please use a supported csv type')

    return filename


def load_raw_data(filename: str) -> RawData:
    """Reads the data (.csv) file written by data acquisition.

    Parameters
    ----------
    - filename : path to the serialized data (csv file)

    Returns
    -------
    RawData object with data held in memory
    """
    return RawData.load(filename)


def load_txt_data() -> str:
    filename = ask_filename('*.txt')  # show dialog box and return the path
    file_name = filename.split('/')[-1]

    if 'txt' not in file_name:
        raise Exception(
            'File type unrecognized. Please use a supported text type')

    return filename


def load_users(data_save_loc) -> List[str]:
    """Load Users.

    Loads user directory names below experiments from the data path defined and returns them as a list.
    If the save data directory is not found, this method returns an empty list assuming no experiments
    have been run yet.
    """
    # build a saved users list, pull out the data save location from parameters
    saved_users = []

    # check the directory is valid, if it is, set path as data save location
    if os.path.isdir(data_save_loc):
        path = data_save_loc

    # check the directory is valid after adding bcipy, if it is, set path as data save location
    elif os.path.isdir(f'{ROOT}/{data_save_loc}'):
        path = f'{ROOT}/{data_save_loc}'

    else:
        log.info(f'User save data location not found at [{data_save_loc}]! Returning empty user list.')
        return saved_users

    # grab all experiments in the directory and iterate over them to get the users
    experiments = fast_scandir(path, return_path=True)

    for experiment in experiments:
        users = fast_scandir(experiment, return_path=False)
        # If it is a new user, append it to the saved_user list
        for user in users:
            if user not in saved_users:
                saved_users.append(user)

    return saved_users


def fast_scandir(directory_name: str, return_path: bool = True) -> List[str]:
    """Fast Scan Directory.

    directory_name: name of the directory to be scanned
    return_path: whether or not to return the scanned directories as a relative path or name.
        False will return the directory name only.
    """
    if return_path:
        return [f.path for f in os.scandir(directory_name) if f.is_dir()]

    return [f.name for f in os.scandir(directory_name) if f.is_dir()]



"""GUI alert messages"""
# pylint: disable=no-name-in-module
import sys
from PyQt5.QtWidgets import QApplication
from bcipy.gui.main import alert_message, AlertMessageType, AlertResponse, AlertMessageResponse


def confirm(message: str) -> bool:
    """Confirmation dialog which allows the user to select between a true and false.

    Parameters
    ----------
        message - alert to display
    Returns
    -------
        users selection : True for selecting Ok, False for Cancel.
    """
    app = QApplication(sys.argv)
    dialog = alert_message(message,
                           message_type=AlertMessageType.INFO,
                           message_response=AlertMessageResponse.OCE)
    button = dialog.exec()

    result = bool(button == AlertResponse.OK.value)
    app.quit()
    return result



"""Functionality related to clocks to keep track of time in experiments."""

from typing import Callable
from pylsl import local_clock


# Adapted from psychopy.
# pylint: disable=invalid-name
class Clock():
    """Monotonic clock that provides high resolution timestamps.

    Parameters
    ----------
        start_at_zero : If `True` the value returned by `getTime` is
            relative to when the clock was started or reset. Default is `False`
        get_time : optional function called to get the current time. The
            default value is the LabStreamingLayer (LSL) local_clock. This is
            the clock used by the acquisition module when sampling data and
            allows us to make time comparisons without conversion.
    """

    def __init__(self,
                 start_at_zero: bool = False,
                 get_time: Callable[[], float] = local_clock):
        """Initialize the clock."""
        self.start_at_zero = start_at_zero
        self._get_time = get_time
        self._last_reset_time = self._get_time()

    def getTime(self) -> float:
        """Returns the current time on this clock in seconds (sub-ms precision).

        This value is whatever the underlying clock uses as its base time but
        is system dependent. e.g. can be time since reboot, time since Unix
        Epoch etc.

        If the start_at_zero property is true, the returned value will be the
        time since the clock was created or reset.
        """
        if self.start_at_zero:
            return self._get_time() - self._last_reset_time
        return self._get_time()

    def reset(self) -> None:
        """Resets the clock."""
        self._last_reset_time = self._get_time()
